import logging
import os
import re
import urllib.parse
from typing import Set, List, Iterable, Dict, Optional, Generator, Tuple

import requests

from bauh.api.http import HttpClient
from bauh.gems.arch import AUR_INDEX_FILE, git
from bauh.gems.arch.exceptions import PackageNotFoundException

URL_INFO = 'https://aur.archlinux.org/rpc/?v=5&type=info&'
URL_SRC_INFO = 'https://aur.archlinux.org/cgit/aur.git/plain/.SRCINFO?h='
URL_SEARCH = 'https://aur.archlinux.org/rpc/?v=5&type=search&arg='
URL_INDEX = 'https://aur.archlinux.org/packages.gz'

RE_SRCINFO_KEYS = re.compile(r'(\w+)\s+=\s+(.+)\n')
RE_SPLIT_DEP = re.compile(r'[<>]?=')

KNOWN_LIST_FIELDS = ('validpgpkeys',
                     'checkdepends',
                     'checkdepends_x86_64',
                     'checkdepends_i686',
                     'depends',
                     'depends_x86_64',
                     'depends_i686',
                     'optdepends',
                     'optdepends_x86_64',
                     'optdepends_i686',
                     'sha256sums',
                     'sha256sums_x86_64',
                     'sha512sums',
                     'sha512sums_x86_64',
                     'source',
                     'source_x86_64',
                     'source_i686',
                     'makedepends',
                     'makedepends_x86_64',
                     'makedepends_i686',
                     'provides',
                     'conflicts')


def map_srcinfo(string: str, pkgname: Optional[str], fields: Set[str] = None) -> dict:
    subinfos, subinfo = [], {}

    key_fields = {'pkgname', 'pkgbase'}

    for field in RE_SRCINFO_KEYS.findall(string):
        key = field[0].strip()
        val = field[1].strip()

        if subinfo and key in key_fields:
            subinfos.append(subinfo)
            subinfo = {key: val}
        elif not fields or key in fields:
            if key not in subinfo:
                subinfo[key] = {val} if key in KNOWN_LIST_FIELDS else val
            else:
                if not isinstance(subinfo[key], set):
                    subinfo[key] = {subinfo[key]}

                subinfo[key].add(val)

    if subinfo:
        subinfos.append(subinfo)

    pkgnames = {s['pkgname'] for s in subinfos if 'pkgname' in s}
    return merge_subinfos(subinfos=subinfos,
                          pkgname=None if (not pkgname or len(pkgnames) == 1 or pkgname not in pkgnames) else pkgname,
                          fields=fields)


def merge_subinfos(subinfos: List[dict], pkgname: Optional[str] = None, fields: Optional[Set[str]] = None) -> dict:
    info = {}
    for subinfo in subinfos:
        if not pkgname or subinfo.get('pkgname') in {None, pkgname}:
            for key, val in subinfo.items():
                if not fields or key in fields:
                    current_val = info.get(key)

                    if current_val is None:
                        info[key] = val
                    else:
                        if not isinstance(current_val, set):
                            current_val = {current_val}
                            info[key] = current_val

                        if isinstance(val, set):
                            current_val.update(val)
                        else:
                            current_val.add(val)

    for field in info.keys():
        val = info.get(field)

        if isinstance(val, set):
            info[field] = [*val]

    return info


class AURClient:

    def __init__(self, http_client: HttpClient, logger: logging.Logger, x86_64: bool):
        self.http_client = http_client
        self.logger = logger
        self.x86_64 = x86_64
        self.srcinfo_cache = {}

    def search(self, words: str) -> dict:
        return self.http_client.get_json(URL_SEARCH + words)

    def get_info(self, names: Iterable[str]) -> Optional[List[dict]]:
        if names:
            try:
                res = self.http_client.get_json(URL_INFO + self._map_names_as_queries(names))
            except requests.exceptions.ConnectionError:
                self.logger.warning('Could not retrieve installed AUR packages API data. It seems the internet connection is off.')
                return

            if res is None:
                self.logger.warning("Call to AUR API's info endpoint has failed")
                return

            error = res.get('error')

            if error:
                self.logger.warning(f"AUR API's info endpoint returned an unexpected error: {error}")
                return

            results = res.get('results')

            if results is not None:
                return results

            self.logger.warning(f"AUR API's info endpoint returned an unexpected response: {res}")

    def map_provided(self, pkgname: str, pkgver: str, provided: Optional[Iterable[str]] = None, strip_epoch: bool = True) -> Set[str]:
        all_provided = {pkgname, f"{pkgname}={pkgver.split('-')[0] if strip_epoch else pkgver}"}

        if provided:
            for provided in provided:
                all_provided.add(provided)
                all_provided.add(provided.split('=', 1)[0])

        return all_provided

    def gen_updates_data(self, names: Iterable[str]) -> Generator[Tuple[str, dict], None, None]:
        pkgs_info = self.get_info(names)

        if pkgs_info:
            for package in pkgs_info:
                pkgname, pkgver = package['Name'], package['Version'].split('-')[0]

                deps = set()

                for dtype in ('Depends', 'MakeDepends', 'CheckDepends'):
                    dep_set = package.get(dtype)
                    if dep_set:
                        deps.update(dep_set)

                conflicts = set()
                pkg_conflicts = package.get('Conflicts')

                if pkg_conflicts:
                    conflicts.update(pkg_conflicts)

                yield pkgname, {
                    'v': pkgver,
                    'b': package.get('PackageBase', pkgname),
                    'r': 'aur',
                    'p': self.map_provided(pkgname=pkgname, pkgver=pkgver, provided=package.get('Provides'), strip_epoch=False),
                    'd': deps,
                    'c': conflicts,
                    'ds': None,
                    's': None}

    def get_src_info(self, name: str, real_name: Optional[str] = None) -> dict:
        srcinfo = self.srcinfo_cache.get(name)

        if srcinfo:
            return srcinfo

        res = self.http_client.get(URL_SRC_INFO + urllib.parse.quote(name))

        if res and res.text:
            srcinfo = map_srcinfo(string=res.text, pkgname=real_name if real_name else name)

            if srcinfo:
                self.srcinfo_cache[name] = srcinfo

            return srcinfo

        self.logger.warning('No .SRCINFO found for {}'.format(name))
        self.logger.info('Checking if {} is based on another package'.format(name))
        # if was not found, it may be based on another package.
        infos = self.get_info((name,))

        if infos:
            info = infos[0]

            info_name = info.get('Name')
            info_base = info.get('PackageBase')
            if info_name and info_base and info_name != info_base:
                self.logger.info('{p} is based on {b}. Retrieving {b} .SRCINFO'.format(p=info_name, b=info_base))
                srcinfo = self.get_src_info(name=info_base, real_name=info_name)

                if srcinfo:
                    self.srcinfo_cache[name] = srcinfo

                return srcinfo

    def extract_required_dependencies(self, srcinfo: dict) -> Set[str]:
        deps = set()

        for attr in ('makedepends',
                     'makedepends_{}'.format('x86_64' if self.x86_64 else 'i686'),
                     'depends',
                     'depends_{}'.format('x86_64' if self.x86_64 else 'i686'),
                     'checkdepends',
                     'checkdepends_{}'.format('x86_64' if self.x86_64 else 'i686')):

            if srcinfo.get(attr):
                deps.update(srcinfo[attr])

        return deps

    def get_required_dependencies(self, name: str) -> Set[str]:
        info = self.get_src_info(name)

        if not info:
            raise PackageNotFoundException(name)

        return self.extract_required_dependencies(info)

    def _map_names_as_queries(self, names: Iterable[str]) -> str:
        return '&'.join((f'arg[]={urllib.parse.quote(n)}' for n in names))

    def read_local_index(self) -> dict:
        self.logger.info('Checking if the cached AUR index file exists')
        if os.path.exists(AUR_INDEX_FILE):
            self.logger.info('Reading AUR index file from {}'.format(AUR_INDEX_FILE))
            index = {}
            with open(AUR_INDEX_FILE) as f:
                for l in f.readlines():
                    if l:
                        lsplit = l.split('=')
                        index[lsplit[0]] = lsplit[1].strip()
            self.logger.info("AUR index file read")
            return index
        self.logger.warning('The AUR index file was not found')

    def download_names(self) -> Set[str]:
        self.logger.info('Downloading AUR index')
        try:
            res = self.http_client.get(URL_INDEX)

            if res and res.text:
                return {n.strip() for n in res.text.split('\n') if n and not n.startswith('#')}
            else:
                self.logger.warning('No data returned from: {}'.format(URL_INDEX))
        except requests.exceptions.ConnectionError:
            self.logger.warning('No internet connection: could not pre-index packages')

        self.logger.info("Finished")

    def read_index(self) -> Iterable[str]:
        try:
            index = self.read_local_index()

            if not index:
                self.logger.warning("Cached AUR index file not found")
                pkgnames = self.download_names()

                if pkgnames:
                    return pkgnames
                else:
                    self.logger.warning("Could not load AUR index on the context")
                    return set()
            else:
                return index.values()
        except:
            return set()

    def clean_caches(self):
        self.srcinfo_cache.clear()

    def map_update_data(self, pkgname: str, latest_version: Optional[str] = None, srcinfo: Optional[dict] = None) -> dict:
        info = self.get_src_info(pkgname) if not srcinfo else srcinfo

        provided = set()
        provided.add(pkgname)

        if info:
            provided.add('{}={}'.format(pkgname, info['pkgver']))
            if info.get('provides'):
                provided.update(info.get('provides'))

            return {'c': info.get('conflicts'), 's': None, 'p': provided, 'r': 'aur',
                    'v': info['pkgver'], 'd': self.extract_required_dependencies(info),
                    'b': info.get('pkgbase', pkgname)}
        else:
            if latest_version:
                provided.add('{}={}'.format(pkgname, latest_version))

            return {'c': None, 's': None, 'p': provided, 'r': 'aur', 'v': latest_version, 'd': set(), 'b': pkgname}

    def fill_update_data(self, output: Dict[str, dict], pkgname: str, latest_version: str, srcinfo: dict = None):
        data = self.map_update_data(pkgname=pkgname, latest_version=latest_version, srcinfo=srcinfo)
        output[pkgname] = data


def is_supported(arch_config: dict) -> bool:
    return arch_config['aur'] and git.is_installed()




import glob
import os
from pathlib import Path
from typing import List, Optional

from bauh import __app_name__
from bauh.api import user
from bauh.api.abstract.model import SoftwarePackage
from bauh.api.paths import AUTOSTART_DIR
from bauh.commons import resource
from bauh.gems.web import ROOT_DIR


class WebApplication(SoftwarePackage):

    def __init__(self, id: Optional[str] = None, url: Optional[str] = None, name: Optional[str] = None, description: Optional[str] = None,
                 icon_url: Optional[str] = None, installation_dir: Optional[str] = None, desktop_entry: Optional[str] = None,
                 installed: bool = False, version: Optional[str] = None, categories: Optional[List[str]] = None,
                 custom_icon: Optional[str] = None, preset_options: Optional[List[str]] = None, save_icon: bool = True,
                 options_set: Optional[List[str]] = None, package_name: Optional[str] = None, source_url: Optional[str] = None,
                 user_agent: Optional[str] = None):
        super(WebApplication, self).__init__(id=id if id else url, name=name, description=description,
                                             icon_url=icon_url, installed=installed, version=version,
                                             categories=categories)
        self.url = url
        self.source_url = source_url if source_url else url
        self.installation_dir = installation_dir
        self.desktop_entry = desktop_entry
        self.preset_options = preset_options
        self.save_icon = save_icon  # if the icon_url should be used instead of the one retrieved by nativefier
        self.options_set = options_set
        self.package_name = package_name
        self.custom_icon = custom_icon
        self.set_custom_icon(custom_icon)
        self.user_agent = user_agent

    def get_source_url(self):
        if self.source_url:
            return self.source_url

        return self.url

    def set_version(self, version: str):
        self.version = str(version) if version else None
        self.latest_version = version

    def has_history(self):
        return False

    def has_info(self):
        return True

    @staticmethod
    def _get_cached_attrs() -> tuple:
        return 'id', 'name', 'version', 'url', 'description', 'icon_url', 'installation_dir', \
               'desktop_entry', 'categories', 'custom_icon', 'options_set', 'save_icon', 'package_name', 'source_url', \
               'user_agent'

    def can_be_downgraded(self):
        return False

    def get_exec_path(self) -> str:
        if self.installation_dir:
            return f'{self.installation_dir}/{self.id}'

    def get_command(self) -> str:
        if self.installation_dir:
            return f"{self.get_exec_path()}{' --no-sandbox' if user.is_root() else ''}"

    def get_type(self):
        return 'web'

    def get_type_icon_path(self) -> str:
        return self.get_default_icon_path()

    def get_default_icon_path(self) -> str:
        return resource.get_path('img/web.svg', ROOT_DIR)

    def get_disk_data_path(self) -> str:
        return f'{self.get_disk_cache_path()}/data.yml'

    def get_disk_icon_path(self) -> str:
        if self.custom_icon:
            return self.custom_icon

        if self.installation_dir:
            return f'{self.installation_dir}/resources/app/icon.png'

    def is_application(self):
        return True

    def supports_disk_cache(self):
        return self.installed

    def get_disk_cache_path(self):
        return self.installation_dir

    def get_data_to_cache(self) -> dict:
        data = {}

        for attr in self._get_cached_attrs():
            if hasattr(self, attr):
                val = getattr(self, attr)

                if val is not None:
                    data[attr] = val

        return data

    def fill_cached_data(self, data: dict):
        for attr in self._get_cached_attrs():
            val = data.get(attr)

            if val and hasattr(self, attr):
                setattr(self, attr, val)

        self.set_custom_icon(self.custom_icon)

    def can_be_run(self) -> bool:
        return self.installed and bool(self.installation_dir)

    def is_trustable(self) -> bool:
        return False

    def get_publisher(self) -> str:
        return __app_name__

    def has_screenshots(self) -> bool:
        return False

    def get_autostart_path(self) -> str:
        if self.desktop_entry:
            return f"{AUTOSTART_DIR}/{self.desktop_entry.split('/')[-1]}"

    def set_custom_icon(self, custom_icon: str):
        self.custom_icon = custom_icon

        if custom_icon:
            self.icon_url = custom_icon

    def get_config_dir(self) -> str:
        if self.package_name:
            config_dir = f'{Path.home()}/.config/{self.package_name}'

            if os.path.exists(config_dir):
                return config_dir
        else:
            if self.installation_dir:
                config_path = f'{Path.home()}/.config'

                if os.path.exists(config_path):
                    config_dirs = glob.glob(f"{config_path}/{self.installation_dir.split('/')[-1]}-nativefier-*")

                    if config_dirs:
                        return config_dirs[0]

    def supports_backup(self) -> bool:
        return False

    def __eq__(self, other):
        if isinstance(other, WebApplication):
            return self.name == other.name and self.url == other.url



from unittest import TestCase
from unittest.mock import Mock, patch

from bauh.gems.web.controller import DEFAULT_LANGUAGE_HEADER
from bauh.gems.web.controller import WebApplicationManager


class ControllerTest(TestCase):

    def test_DEFAULT_LANGUAGE_HEADER(self):
        self.assertEqual('en-US, en', DEFAULT_LANGUAGE_HEADER)


class WebApplicationManagerTest(TestCase):

    def setUp(self):
        self.manager = WebApplicationManager(context=Mock())

    @patch('locale.getdefaultlocale', side_effect=Exception)
    def test_get_accept_language_header__must_return_default_locale_when_exception_raised(self, getdefaultlocale: Mock):
        returned = self.manager.get_accept_language_header()
        self.assertEqual(DEFAULT_LANGUAGE_HEADER, returned)
        getdefaultlocale.assert_called_once()

    @patch('locale.getdefaultlocale', return_value=None)
    def test_get_accept_language_header__must_return_default_locale_when_no_locale_is_returned(self, getdefaultlocale: Mock):
        returned = self.manager.get_accept_language_header()
        self.assertEqual(DEFAULT_LANGUAGE_HEADER, returned)
        getdefaultlocale.assert_called_once()

    @patch('locale.getdefaultlocale', return_value=['es_AR'])
    def test_get_accept_language_header__must_return_the_system_locale_without_underscore_plus_default_locale(self, getdefaultlocale: Mock):
        returned = self.manager.get_accept_language_header()
        self.assertEqual(f'es-AR, es, {DEFAULT_LANGUAGE_HEADER}', returned)
        getdefaultlocale.assert_called_once()

    @patch('locale.getdefaultlocale', return_value=['es'])
    def test_get_accept_language_header__must_return_the_simple_system_locale_plus_default_locale(self, getdefaultlocale: Mock):
        returned = self.manager.get_accept_language_header()
        self.assertEqual(f'es, {DEFAULT_LANGUAGE_HEADER}', returned)
        getdefaultlocale.assert_called_once()

    @patch('locale.getdefaultlocale', return_value=['en_IN'])
    def test_get_accept_language_header__must_not_concatenate_default_locale_if_system_locale_has_it(self, getdefaultlocale: Mock):
        returned = self.manager.get_accept_language_header()
        self.assertEqual(f'en-IN, en', returned)
        getdefaultlocale.assert_called_once()



import re
import traceback
from logging import Logger
from threading import Thread
from typing import Set, List, Tuple, Dict, Iterable, Optional, Generator

from bauh.api.abstract.handler import ProcessWatcher
from bauh.gems.arch import pacman, message, sorting, confirmation
from bauh.gems.arch.aur import AURClient
from bauh.gems.arch.exceptions import PackageNotFoundException
from bauh.gems.arch.version import match_required_version
from bauh.view.util.translation import I18n


class DependenciesAnalyser:

    def __init__(self, aur_client: AURClient, i18n: I18n, logger: Logger):
        self.aur_client = aur_client
        self.i18n = i18n
        self.re_dep_operator = re.compile(r'([<>=]+)')
        self._log = logger

    def _fill_repository(self, name: str, output: List[Tuple[str, str]]):

        repository = pacman.read_repository_from_info(name)

        if repository:
            output.append((name, repository))
            return

        guess = pacman.guess_repository(name)

        if guess:
            output.append(guess)
            return

        aur_info = self.aur_client.get_src_info(name)

        if aur_info:
            output.append((name, 'aur'))
            return

        output.append((name, ''))

    def get_missing_packages(self, names: Set[str], repository: Optional[str] = None, in_analysis: Optional[Set[str]] = None) -> \
            List[Tuple[str, str]]:
        """
        :param names:
        :param repository:
        :param in_analysis: global set storing all names in analysis to avoid repeated recursion
        :return:
        """
        global_in_analysis = in_analysis if in_analysis else set()

        missing_names = pacman.check_missing({n for n in names if n not in in_analysis})

        if missing_names:
            missing_root = []
            threads = []

            if not repository:
                for name in missing_names:
                    t = Thread(target=self._fill_repository, args=(name, missing_root))
                    t.start()
                    threads.append(t)

                for t in threads:
                    t.join()

                threads.clear()

                # checking if there is any unknown dependency:
                for rdep in missing_root:
                    if not rdep[1]:
                        return missing_root
                    else:
                        global_in_analysis.add(rdep[0])
            else:
                for missing in missing_names:
                    missing_root.append((missing, repository))
                    global_in_analysis.add(missing)

            missing_sub = []
            for rdep in missing_root:
                subdeps = self.aur_client.get_required_dependencies(rdep[0]) if rdep[1] == 'aur' else pacman.read_dependencies(rdep[0])
                subdeps_not_analysis = {sd for sd in subdeps if sd not in global_in_analysis}

                if subdeps_not_analysis:
                    missing_subdeps = self.get_missing_packages(subdeps_not_analysis, in_analysis=global_in_analysis)

                    # checking if there is any unknown:
                    if missing_subdeps:
                        for subdep in missing_subdeps:
                            if not subdep[0]:
                                return [*missing_subdeps, *missing_root]

                            if subdep[0] not in missing_names:
                                missing_sub.append(subdep)
            return [*missing_sub, *missing_root]

    def get_missing_subdeps_of(self, names: Set[str], repository: str) -> List[Tuple[str, str]]:
        missing = []
        already_added = {*names}
        in_analyses = {*names}

        for name in names:
            subdeps = self.aur_client.get_required_dependencies(name) if repository == 'aur' else pacman.read_dependencies(name)

            if subdeps:
                missing_subdeps = self.get_missing_packages(subdeps, in_analysis=in_analyses)

                if missing_subdeps:
                    for subdep in missing_subdeps:  # checking if there is any unknown:
                        if subdep[0] not in already_added:
                            missing.append(subdep)

                        if not subdep[0]:
                            return missing
        return missing

    def get_missing_subdeps(self, name: str, repository: str, srcinfo: dict = None) -> List[Tuple[str, str]]:
        missing = []
        already_added = {name}
        in_analyses = {name}

        if repository == 'aur':
            subdeps = self.aur_client.get_required_dependencies(name) if not srcinfo else self.aur_client.extract_required_dependencies(srcinfo)
        else:
            subdeps = pacman.read_dependencies(name)

        if subdeps:
            missing_subdeps = self.get_missing_packages(subdeps, in_analysis=in_analyses)

            if missing_subdeps:
                for subdep in missing_subdeps:  # checking if there is any unknown:
                    if subdep[0] not in already_added:
                        missing.append(subdep)

                    if not subdep[0]:
                        return missing
        return missing

    def map_known_missing_deps(self, known_deps: Dict[str, str], watcher: ProcessWatcher, check_subdeps: bool = True) -> Optional[List[Tuple[str, str]]]:
        sorted_deps = []  # it will hold the proper order to install the missing dependencies

        repo_deps, aur_deps = set(), set()

        for dep, repo in known_deps.items():
            if repo == 'aur':
                aur_deps.add(dep)
            else:
                repo_deps.add(dep)

        if check_subdeps:
            for deps in ((repo_deps, 'repo'), (aur_deps, 'aur')):
                if deps[0]:
                    missing_subdeps = self.get_missing_subdeps_of(deps[0], deps[1])

                    if missing_subdeps:
                        for dep in missing_subdeps:
                            if not dep[1]:
                                dependents = ', '.join(deps[0])  # it is not possible to know which is the exact dependent
                                message.show_dep_not_found(dep[0], self.i18n, watcher, dependent=dependents)
                                return

                        for dep in missing_subdeps:
                            if dep not in sorted_deps:
                                sorted_deps.append(dep)

        for dep, repo in known_deps.items():
            if repo != 'aur':
                data = (dep, repo)
                if data not in sorted_deps:
                    sorted_deps.append(data)

        for dep in aur_deps:
            sorted_deps.append((dep, 'aur'))

        return sorted_deps

    def _find_repo_providers(self, dep_name: str, dep_exp: str, remote_provided_map: Dict[str, Set[str]],
                             deps_data: Dict[str, dict], remote_repo_map: Dict[str, str]) -> Generator[Tuple[str, str, Optional[dict]], None, None]:
        if dep_name == dep_exp:
            providers = remote_provided_map.get(dep_name)

            if providers:
                for pkgname in providers:
                    yield pkgname, remote_repo_map.get(pkgname), None

            else:  # try to find the package through the pacman's search mechanism
                match = pacman.find_one_match(dep_name)

                if match:
                    yield match, remote_repo_map.get(match), None

        else:  # handling cases when the dep has an expression ( e.g: xpto>=0.12 )
            exact_exp_providers = remote_provided_map.get(dep_exp)

            if exact_exp_providers:
                for p in exact_exp_providers:
                    yield p, remote_repo_map.get(p), None
            else:
                providers = remote_provided_map.get(dep_name)

                if not providers:  # try to find the package through the pacman's search mechanism
                    match = pacman.find_one_match(dep_name)

                    if match:
                        providers = {match}

                if providers:
                    providers_no_provided_data = {p for p in providers if p not in deps_data}
                    missing_providers_data = None

                    if providers_no_provided_data:
                        missing_providers_data = pacman.map_updates_data(providers_no_provided_data)

                        if not missing_providers_data:
                            raise Exception(f"Could not retrieve information from providers: "
                                            f"{', '.join(providers_no_provided_data)}")

                        data_not_found = {p for p in providers if p not in missing_providers_data}

                        if data_not_found:
                            raise Exception(f"Could not retrieve information from providers: "
                                            f"{', '.join(data_not_found)}")

                    split_informed_dep = self.re_dep_operator.split(dep_exp)

                    version_required = split_informed_dep[2]
                    exp_op = split_informed_dep[1] if split_informed_dep[1] != '=' else '=='

                    for p in providers:
                        info = deps_data.get(p)

                        if not info and missing_providers_data:
                            info = missing_providers_data[p]

                        for provided_exp in info['p']:
                            split_dep = self.re_dep_operator.split(provided_exp)

                            if len(split_dep) == 3 and split_dep[0] == dep_name:
                                version_provided = split_dep[2]

                                if match_required_version(version_provided, exp_op, version_required):
                                    yield p, remote_repo_map.get(p), info
                                    break

    def _find_aur_providers(self, dep_name: str, dep_exp: str, aur_index: Iterable[str], exact_match: bool) -> Generator[Tuple[str, dict], None, None]:
        if exact_match and dep_name in aur_index:
            if dep_name == dep_exp:
                yield from self.aur_client.gen_updates_data((dep_name,))
                return
            else:
                for _, dep_data in self.aur_client.gen_updates_data((dep_name,)):
                    split_informed_dep = self.re_dep_operator.split(dep_exp)
                    version_required = split_informed_dep[2]
                    exp_op = split_informed_dep[1].strip()

                    if match_required_version(dep_data['v'], exp_op, version_required):
                        yield dep_name, dep_data
                        return

        aur_search = self.aur_client.search(dep_name)

        if aur_search:
            aur_results = aur_search.get('results')

            if aur_results:
                if dep_name == dep_exp:
                    version_required, exp_op = None, None
                else:
                    split_informed_dep = self.re_dep_operator.split(dep_exp)
                    version_required = split_informed_dep[2]
                    exp_op = split_informed_dep[1] if split_informed_dep[1] != '=' else '=='

                for pkgname, pkgdata in self.aur_client.gen_updates_data((aur_res['Name'] for aur_res in aur_results)):
                    if pkgname == dep_name or (dep_name in pkgdata['p']):
                        try:
                            if not version_required or match_required_version(pkgdata['v'], exp_op,
                                                                              version_required):
                                yield pkgname, pkgdata
                        except:
                            self._log.warning(f"Could not compare AUR package '{pkgname}' version '{pkgdata['v']}' "
                                              f"with the dependency expression '{dep_exp}'")
                            traceback.print_exc()

    def _fill_missing_dep(self, dep_name: str, dep_exp: str, aur_index: Iterable[str],
                          missing_deps: Set[Tuple[str, str]],
                          remote_provided_map: Dict[str, Set[str]], remote_repo_map: Dict[str, str],
                          repo_deps: Set[str], aur_deps: Set[str], deps_data: Dict[str, dict], watcher: ProcessWatcher,
                          automatch_providers: bool, prefer_repository_provider: bool, dependent: Optional[str] = None):

        repo_matches = None

        for pkgname, repo, data in self._find_repo_providers(dep_name=dep_name, dep_exp=dep_exp,
                                                             remote_repo_map=remote_repo_map,
                                                             remote_provided_map=remote_provided_map,
                                                             deps_data=deps_data):
            if automatch_providers and pkgname == dep_name:
                missing_deps.add((pkgname, repo))
                repo_deps.add(pkgname)

                if data:
                    deps_data[pkgname] = data

                return

            if repo_matches is None:
                repo_matches = []

            repo_matches.append((pkgname, repo, data))

        if prefer_repository_provider and repo_matches and len(repo_matches) == 1:
            pkg_name, pkg_repo, pkg_data = repo_matches[0]
            missing_deps.add((pkg_name, pkg_repo))
            repo_deps.add(pkg_name)

            if pkg_data:
                deps_data[pkg_name] = pkg_data

            return

        aur_matches = None

        if aur_index:
            for pkgname, pkgdata in self._find_aur_providers(dep_name=dep_name, dep_exp=dep_exp, aur_index=aur_index,
                                                             exact_match=automatch_providers):
                if automatch_providers and pkgname == dep_name:
                    missing_deps.add((pkgname, 'aur'))
                    aur_deps.add(pkgname)
                    deps_data[pkgname] = pkgdata
                    return

                if aur_matches is None:
                    aur_matches = []

                aur_matches.append((pkgname, pkgdata))

        total_matches = (len(repo_matches) if repo_matches else 0) + (len(aur_matches) if aur_matches else 0)

        if total_matches == 0:
            self.__raise_dependency_not_found(dep_exp, watcher, dependent)
        elif total_matches == 1:
            if repo_matches:
                repo_pkg = [*repo_matches][0]

                if repo_pkg[2]:
                    deps_data[repo_pkg[0]] = repo_pkg[2]  # already updating the deps data (if available)

                match = (repo_pkg[0], repo_pkg[1])
                repo_deps.add(repo_pkg[0])
            else:
                aur_pkg = aur_matches[0]
                deps_data[aur_pkg[0]] = aur_pkg[1]  # already updating the deps data (if available)
                aur_deps.add(aur_pkg[0])
                match = aur_pkg[0], 'aur'

            missing_deps.add(match)

        elif total_matches > 1:
            missing_deps.add((dep_name, '__several__'))

            if repo_matches:
                repo_deps.add(dep_name)

            if aur_matches:
                aur_deps.add(dep_name)

                for pkgname, _ in aur_matches:
                    for key in (dep_name, dep_exp):
                        key_provided = remote_provided_map.get(key, set())
                        remote_provided_map[key] = key_provided
                        key_provided.add(pkgname)

                    if pkgname not in remote_repo_map:
                        remote_repo_map[pkgname] = 'aur'

    def __raise_dependency_not_found(self, dep_exp: str, watcher: Optional[ProcessWatcher], dependent: Optional[str] = None):
        if watcher:
            message.show_dep_not_found(depname=dep_exp, i18n=self.i18n, watcher=watcher, dependent=dependent)
            raise PackageNotFoundException(dep_exp)
        else:
            raise PackageNotFoundException(dep_exp)

    def _fill_aur_updates_data(self, pkgnames: Iterable[str], output_data: dict):
        if pkgnames:
            for pkgname, pkgdata in self.aur_client.gen_updates_data(pkgnames):
                output_data[pkgname] = pkgdata

    def map_missing_deps(self, pkgs_data: Dict[str, dict], provided_map: Dict[str, Set[str]],
                         remote_provided_map: Dict[str, Set[str]], remote_repo_map: Dict[str, str],
                         aur_index: Iterable[str], deps_checked: Set[str], deps_data: Dict[str, dict],
                         sort: bool, watcher: ProcessWatcher, choose_providers: bool = True,
                         automatch_providers: bool = False, prefer_repository_provider: bool = False) -> Optional[List[Tuple[str, str]]]:
        sorted_deps = []  # it will hold the proper order to install the missing dependencies

        missing_deps, repo_missing, aur_missing = set(), set(), set()

        deps_checked.update(pkgs_data.keys())

        for p, data in pkgs_data.items():
            if data['d']:
                for dep in data['d']:
                    if dep in pkgs_data:
                        continue
                    if dep not in provided_map:
                        dep_split = self.re_dep_operator.split(dep)
                        dep_name = dep_split[0].strip()

                        if dep_name not in deps_checked:
                            deps_checked.add(dep_name)

                            if dep_name not in provided_map:
                                self._fill_missing_dep(dep_name=dep_name, dep_exp=dep, aur_index=aur_index,
                                                       missing_deps=missing_deps,
                                                       remote_provided_map=remote_provided_map,
                                                       remote_repo_map=remote_repo_map,
                                                       repo_deps=repo_missing, aur_deps=aur_missing, watcher=watcher,
                                                       deps_data=deps_data,
                                                       automatch_providers=automatch_providers,
                                                       prefer_repository_provider=prefer_repository_provider,
                                                       dependent=p)
                            else:
                                version_pattern = '{}='.format(dep_name)
                                version_found = [p for p in provided_map if p.startswith(version_pattern)]

                                if version_found:
                                    version_found = version_found[0].split('=')[1]
                                    version_required = dep_split[2]
                                    op = dep_split[1].strip()

                                    if not match_required_version(version_found, op, version_required):
                                        self._fill_missing_dep(dep_name=dep_name, dep_exp=dep, aur_index=aur_index,
                                                               missing_deps=missing_deps,
                                                               remote_provided_map=remote_provided_map,
                                                               remote_repo_map=remote_repo_map,
                                                               repo_deps=repo_missing, aur_deps=aur_missing,
                                                               watcher=watcher,
                                                               deps_data=deps_data,
                                                               automatch_providers=automatch_providers,
                                                               prefer_repository_provider=prefer_repository_provider,
                                                               dependent=p)
                                else:
                                    self._fill_missing_dep(dep_name=dep_name, dep_exp=dep, aur_index=aur_index,
                                                           missing_deps=missing_deps,
                                                           remote_provided_map=remote_provided_map,
                                                           remote_repo_map=remote_repo_map,
                                                           repo_deps=repo_missing, aur_deps=aur_missing,
                                                           watcher=watcher,
                                                           deps_data=deps_data,
                                                           automatch_providers=automatch_providers,
                                                           prefer_repository_provider=prefer_repository_provider,
                                                           dependent=p)

        if missing_deps:
            self._fill_single_providers_data(missing_deps, repo_missing, aur_missing, deps_data)

            missing_subdeps = self.map_missing_deps(pkgs_data={**deps_data}, provided_map=provided_map, aur_index=aur_index,
                                                    deps_checked=deps_checked, sort=False, deps_data=deps_data,
                                                    watcher=watcher,
                                                    remote_provided_map=remote_provided_map,
                                                    remote_repo_map=remote_repo_map,
                                                    automatch_providers=automatch_providers,
                                                    choose_providers=False,
                                                    prefer_repository_provider=prefer_repository_provider)

            if missing_subdeps:
                missing_deps.update(missing_subdeps)

        if sort:
            sorted_deps.extend(sorting.sort(deps_data.keys(), deps_data))
        else:
            sorted_deps.extend(((dep[0], dep[1]) for dep in missing_deps))

        if sorted_deps and choose_providers:
            return self.fill_providers_deps(missing_deps=sorted_deps, provided_map=provided_map,
                                            remote_provided_map=remote_provided_map, remote_repo_map=remote_repo_map,
                                            watcher=watcher, sort=sort, already_checked=deps_checked,
                                            aur_idx=aur_index, deps_data=deps_data,
                                            automatch_providers=automatch_providers,
                                            prefer_repository_provider=prefer_repository_provider)

        return sorted_deps

    def _fill_single_providers_data(self, all_missing_deps: Iterable[Tuple[str, str]], repo_missing_deps: Iterable[str], aur_missing_deps: Iterable[str], deps_data: Dict[str, dict]):
        """
            fills the missing data of the single dependency providers since they are already considered dependencies
            (when several providers are available for given a dependency, the user must choose first)
        """
        repo_providers_no_data, aur_providers_no_data = None, None

        for dep_name, dep_repo in all_missing_deps:
            if dep_repo == '__several__':
                deps_data[dep_name] = {'d': None, 'p': {dep_name}, 'r': dep_repo}
            elif dep_name not in deps_data:
                if repo_missing_deps and dep_name in repo_missing_deps:
                    if repo_providers_no_data is None:
                        repo_providers_no_data = set()

                    repo_providers_no_data.add(dep_name)
                elif aur_missing_deps and dep_name in aur_missing_deps:
                    if aur_providers_no_data is None:
                        aur_providers_no_data = set()

                    aur_providers_no_data.add(dep_name)

        aur_data_filler, aur_providers_data = None, None

        if aur_providers_no_data:
            aur_providers_data = dict()
            aur_data_filler = Thread(target=self._fill_aur_updates_data,
                                     args=(aur_providers_no_data, aur_providers_data))
            aur_data_filler.start()

        if repo_providers_no_data:
            repo_providers_data = pacman.map_updates_data(repo_providers_no_data)

            if repo_providers_data:
                deps_data.update(repo_providers_data)

        if aur_data_filler:
            aur_data_filler.join()

            if aur_providers_data:
                deps_data.update(aur_providers_data)

    def fill_providers_deps(self, missing_deps: List[Tuple[str, str]],
                            provided_map: Dict[str, Set[str]], remote_repo_map: Dict[str, str],
                            already_checked: Set[str], remote_provided_map: Dict[str, Set[str]],
                            deps_data: Dict[str, dict], aur_idx: Iterable[str], sort: bool,
                            watcher: ProcessWatcher, automatch_providers: bool,
                            prefer_repository_provider: bool) -> Optional[List[Tuple[str, str]]]:
        """
        :param missing_deps:
        :param provided_map:
        :param remote_repo_map:
        :param already_checked:
        :param remote_provided_map:
        :param deps_data:
        :param aur_idx:
        :param sort:
        :param watcher:
        :param automatch_providers
        :param prefer_repository_provider
        :return: all deps sorted or None if the user declined the providers options
        """

        deps_providers = map_providers({data[0] for data in missing_deps if data[1] == '__several__'},
                                       remote_provided_map)

        if deps_providers:
            providers_repos = {}
            repos_providers = set()

            for providers in deps_providers.values():
                for provider in providers:
                    if remote_repo_map.get(provider) == 'aur':
                        providers_repos[provider] = 'aur'
                    else:
                        repos_providers.add(provider)

            providers_repos.update(pacman.map_repositories(repos_providers))
            selected_providers = confirmation.request_providers(deps_providers, providers_repos, watcher, self.i18n)

            if not selected_providers:
                return
            else:
                # adding the chosen providers for re-checking the missing dependencies
                repo_selected, aur_selected = set(), set()

                for provider in selected_providers:
                    if provider in repos_providers:
                        repo_selected.add(provider)
                    else:
                        aur_selected.add(provider)

                providers_data = dict()

                if repo_selected:
                    providers_data.update(pacman.map_updates_data(repo_selected))
                    # adding the providers as "installed" packages
                    provided_map.update(pacman.map_provided(remote=True,  pkgs=repo_selected))

                if aur_selected:
                    for pkgname, pkgdata in self.aur_client.gen_updates_data(aur_selected):
                        providers_data[pkgname] = pkgdata
                        for provider in pkgdata['p']:  # adding the providers as "installed" packages
                            currently_provided = provided_map.get(provider, set())
                            provided_map[provider] = currently_provided
                            currently_provided.add(pkgname)

                providers_deps = self.map_missing_deps(pkgs_data=providers_data,
                                                       provided_map=provided_map,
                                                       aur_index=aur_idx,
                                                       deps_checked=already_checked,
                                                       deps_data=deps_data,
                                                       sort=False,
                                                       remote_provided_map=remote_provided_map,
                                                       remote_repo_map=remote_repo_map,
                                                       watcher=watcher,
                                                       choose_providers=True,
                                                       automatch_providers=automatch_providers,
                                                       prefer_repository_provider=prefer_repository_provider)

                if providers_deps is None:  # it means the user called off the installation process
                    return

                # cleaning the already mapped providers deps:
                to_remove = []

                for idx, dep in enumerate(missing_deps):
                    if dep[1] == '__several__':
                        to_remove.append(idx)

                for idx, to_remove in enumerate(to_remove):
                    del missing_deps[to_remove - idx]

                missing_deps.extend(((p, providers_repos.get(p, 'aur')) for p in selected_providers))

                for dep in providers_deps:
                    if dep not in missing_deps and dep[1] != '__several__':
                        missing_deps.append(dep)

                deps_data.update(providers_data)

                if not self.fill_providers_deps(missing_deps=missing_deps, provided_map=provided_map,
                                                remote_repo_map=remote_repo_map, already_checked=already_checked,
                                                aur_idx=aur_idx, remote_provided_map=remote_provided_map,
                                                deps_data=deps_data, sort=False, watcher=watcher,
                                                automatch_providers=automatch_providers,
                                                prefer_repository_provider=prefer_repository_provider):
                    return

                if sort:
                    missing_to_sort = {d[0] for d in missing_deps if d[1] != '__several__'}
                    return sorting.sort(missing_to_sort, deps_data, provided_map)

        return missing_deps

    def map_all_required_by(self, pkgnames: Iterable[str], to_ignore: Set[str]) -> Set[str]:
        if pkgnames:
            to_ignore.update(pkgnames)

        all_requirements = {req for reqs in pacman.map_required_by(pkgnames).values() for req in reqs if req not in to_ignore}

        if all_requirements:
            sub_requirements = self.map_all_required_by(all_requirements, to_ignore)

            if sub_requirements:
                all_requirements.update(sub_requirements)
                return all_requirements

        return all_requirements


def map_providers(pkgs: Iterable[str], remote_provided_map: Dict[str, Set[str]]) -> Dict[str, Set[str]]:
    res = {}

    for p in pkgs:
        providers = remote_provided_map.get(p)

        if providers and len(providers) > 1:
            res[p] = providers

    return res



import re
from io import StringIO
from re import Pattern
from typing import Optional, Iterable, Tuple

from bauh.api.abstract.model import SoftwarePackage, CustomSoftwareAction
from bauh.commons import resource
from bauh.gems.appimage import ROOT_DIR, INSTALLATION_DIR
from bauh.view.util.translation import I18n


class AppImage(SoftwarePackage):

    __actions_local_installation: Optional[Tuple[CustomSoftwareAction, ...]] = None
    __cached_attrs: Optional[Tuple[str, ...]] = None
    __re_many_spaces: Optional[Pattern] = None

    @classmethod
    def actions_local_installation(cls) -> Tuple[CustomSoftwareAction, ...]:
        if cls.__actions_local_installation is None:
            cls.__actions_local_installation = (CustomSoftwareAction(i18n_label_key='appimage.custom_action.manual_update',
                                                                     i18n_status_key='appimage.custom_action.manual_update.status',
                                                                     i18n_description_key='appimage.custom_action.manual_update.desc',
                                                                     manager_method='update_file',
                                                                     requires_root=False,
                                                                     icon_path=resource.get_path('img/refresh.svg', ROOT_DIR),
                                                                     requires_confirmation=False),)

        return cls.__actions_local_installation

    @classmethod
    def cached_attrs(cls) -> Tuple[str, ...]:
        if cls.__cached_attrs is None:
            cls.__cached_attrs = ('name', 'description', 'version', 'url_download', 'author', 'license', 'source',
                                  'icon_path', 'github', 'categories', 'imported', 'install_dir', 'symlink')

        return cls.__cached_attrs

    @classmethod
    def re_many_spaces(cls) -> Pattern:
        if cls.__re_many_spaces is None:
            cls.__re_many_spaces = re.compile(r'\s+')

        return cls.__re_many_spaces

    def __init__(self, name: str = None, description: str = None, github: str = None, source: str = None, version: str = None,
                 url_download: str = None, url_icon: str = None, url_screenshot: str = None, license: str = None, author: str = None,
                 categories=None, icon_path: str = None, installed: bool = False,
                 url_download_latest_version: str = None, local_file_path: str = None, imported: bool = False,
                 i18n: I18n = None, install_dir: str = None, updates_ignored: bool = False,
                 symlink: str = None, **kwargs):
        super(AppImage, self).__init__(id=name, name=name, version=version, latest_version=version,
                                       icon_url=url_icon, license=license, description=description,
                                       installed=installed)
        self.source = source
        self.github = github
        self.categories = (categories.split(',') if isinstance(categories, str) else categories) if categories else None
        self.url_download = url_download
        self.icon_path = icon_path
        self.url_screenshot = url_screenshot
        self.author = author
        self.url_download_latest_version = url_download_latest_version
        self.local_file_path = local_file_path
        self.imported = imported
        self.i18n = i18n
        self.install_dir = install_dir
        self.updates_ignored = updates_ignored
        self.symlink = symlink

    def __repr__(self):
        return "{} (name={}, github={})".format(self.__class__.__name__, self.name, self.github)

    def can_be_installed(self):
        return not self.installed and self.url_download

    def has_history(self):
        return self.installed and not self.imported

    def has_info(self):
        return self.installed if self.imported else True

    def can_be_downgraded(self):
        return self.installed and not self.imported

    def get_type(self):
        return 'AppImage'

    def get_default_icon_path(self):
        return self.get_type_icon_path()

    def get_type_icon_path(self):
        return resource.get_path('img/appimage.svg', ROOT_DIR)

    def is_application(self):
        return True

    def get_data_to_cache(self) -> dict:
        data = {}

        for a in self.cached_attrs():
            val = getattr(self, a)
            if val:
                data[a] = val

        return data

    def fill_cached_data(self, data: dict):
        for a in self.cached_attrs():
            val = data.get(a)

            if val:
                setattr(self, a, val)

    def can_be_run(self) -> bool:
        return self.installed

    def get_publisher(self) -> str:
        return self.author

    def get_disk_cache_path(self) -> str:
        if self.install_dir:
            return self.install_dir
        elif self.name:
            return f'{INSTALLATION_DIR}/{self.name.lower()}'

    def get_disk_icon_path(self):
        return self.icon_path

    def has_screenshots(self):
        return not self.installed and self.url_screenshot

    def get_name_tooltip(self) -> str:
        if self.name and self.imported:
            return '{} ({})'.format(self.name, self.i18n['imported'])

        return self.name

    def get_custom_actions(self) -> Optional[Iterable[CustomSoftwareAction]]:
        if self.installed and self.imported:
            return self.actions_local_installation()

    def supports_backup(self) -> bool:
        return False

    def supports_ignored_updates(self) -> bool:
        return self.installed and not self.imported

    def is_update_ignored(self) -> bool:
        return self.updates_ignored

    def __eq__(self, other):
        if isinstance(other, AppImage):
            return self.name == other.name and self.local_file_path == other.local_file_path

    def get_clean_name(self) -> Optional[str]:
        if self.name:
            return self.re_many_spaces().sub('-', self.name.lower().strip())

    def to_desktop_entry(self) -> str:
        de = StringIO()
        de.write("[Desktop Entry]\nType=Application\nName={}\n".format(self.name))

        if self.description:
            de.write("Comment={}\n".format(self.description.replace('\n', ' ')))

        if self.install_dir and self.local_file_path:
            de.write('Exec="{}/{}"\n'.format(self.install_dir, self.local_file_path.split('/')[-1]))

        if self.icon_path:
            de.write('Icon={}\n'.format(self.icon_path))

        if self.categories:
            de.write('Categories={};\n'.format(';'.join((c for c in self.categories if c.lower() != 'imported'))))

        de.write('Terminal=false')
        de.seek(0)
        return de.read()



from typing import Iterable, Optional

from bauh.api.abstract.handler import ProcessWatcher
from bauh.api.abstract.view import MessageType
from bauh.commons.html import bold
from bauh.view.util.translation import I18n


def show_deps_not_installed(watcher: ProcessWatcher, pkgname: str, depnames: Iterable[str], i18n: I18n):
    deps = ', '.join((bold(d) for d in depnames))
    watcher.show_message(title=i18n['error'].capitalize(),
                         body=i18n['arch.install.dependency.install.error'].format(deps, bold(pkgname)),
                         type_=MessageType.ERROR)


def show_dep_not_found(depname: str, i18n: I18n, watcher: ProcessWatcher, dependent: Optional[str] = None):

    source = f" {bold('(' + dependent + ')')}" if dependent else ''

    body = f"<p>{i18n['arch.install.dep_not_found.body.l1'].format(dep=bold(depname), source=source)}</p>" \
           f"<p><{i18n['arch.install.dep_not_found.body.l2']}</p>" \
           f"<p>{i18n['arch.install.dep_not_found.body.l3']}</p>"

    watcher.show_message(title=i18n['arch.install.dep_not_found.title'].capitalize(),
                         body=body,
                         type_=MessageType.ERROR)


def show_optdeps_not_installed(depnames: Iterable[str], watcher: ProcessWatcher, i18n: I18n):
    deps = ', '.join((bold(d) for d in depnames))
    watcher.show_message(title=i18n['error'].capitalize(),
                         body=i18n['arch.install.optdep.error'].format(deps),
                         type_=MessageType.ERROR)



from packaging.version import Version

from bauh.api.abstract.model import SoftwarePackage, PackageStatus
from bauh.commons import resource
from bauh.gems.flatpak import ROOT_DIR, VERSION_1_2, VERSION_1_5
from bauh.view.util.translation import I18n


class FlatpakApplication(SoftwarePackage):

    def __init__(self, id: str = None, name: str = None, version: str = None, latest_version: str = None,
                 description: str = None, branch: str = None, arch: str = None, origin: str = None,
                 runtime: bool = False, ref: str = None, commit: str = None, installation: str = None,
                 i18n: I18n = None, partial: bool = False, updates_ignored: bool = False,  installed: bool = False,
                 update: bool = False, update_component: bool = False):
        super(FlatpakApplication, self).__init__(id=id, name=name, version=version, latest_version=latest_version,
                                                 description=description, installed=installed, update=update)
        self.ref = ref
        self.branch = branch
        self.arch = arch
        self.origin = origin
        self.runtime = runtime
        self.commit = commit
        self.partial = partial
        self.installation = installation if installation else 'system'
        self.i18n = i18n
        self.base_id = None
        self.base_ref = None
        self.updates_ignored = updates_ignored
        self.update_component = update_component  # if it is a new app/runtime that has come as an update

        if runtime:
            self.categories = ['runtime']

    def is_incomplete(self):
        return self.description is None and self.icon_url

    def has_history(self) -> bool:
        return not self.partial and not self.update_component and self.installed and self.ref

    def has_info(self):
        return bool(self.id)

    def can_be_downgraded(self):
        return not self.partial and not self.update_component and self.installed and self.ref

    def get_type(self):
        return 'flatpak'

    def get_default_icon_path(self):
        return resource.get_path('img/flatpak.svg', ROOT_DIR)

    def get_type_icon_path(self):
        return self.get_default_icon_path()

    def is_application(self):
        return not self.runtime and not self.partial and not self.update_component

    def get_disk_cache_path(self):
        return super(FlatpakApplication, self).get_disk_cache_path() + '/installed/' + self.id

    def get_data_to_cache(self):
        if not self.update_component:
            return {
                'description': self.description,
                'icon_url': self.icon_url,
                'latest_version': self.latest_version,
                'version': self.version,
                'name': self.name,
                'categories': self.categories
            }

    def fill_cached_data(self, data: dict):
        for attr in self.get_data_to_cache().keys():
            if data.get(attr) and not getattr(self, attr):
                setattr(self, attr, data[attr])

    def can_be_run(self) -> bool:
        return self.installed and not self.runtime and not self.partial and not self.update_component

    def get_publisher(self):
        return self.origin

    def gen_partial(self, partial_id: str) -> "FlatpakApplication":
        partial = FlatpakApplication()
        partial.partial = True
        partial.id = partial_id
        partial.base_id = self.id
        partial.installation = self.installation
        partial.origin = self.origin
        partial.branch = self.branch
        partial.i18n = self.i18n
        partial.arch = self.arch
        partial.name = self.name
        partial.version = self.version
        partial.latest_version = self.latest_version
        partial.installed = self.installed
        partial.runtime = True

        if self.ref:
            partial.base_ref = self.ref
            partial.ref = '/'.join((partial_id, *self.ref.split('/')[1:]))
            partial.status = PackageStatus.READY
            partial.name += ' ({})'.format(partial_id.split('.')[-1])

        return partial

    def get_name_tooltip(self) -> str:
        if self.installed and self.installation and self.i18n is not None:
            return '{} ({})'.format(self.name, self.i18n['flatpak.info.installation.{}'.format(self.installation.lower().strip())])

        return self.name

    def supports_backup(self) -> bool:
        return True

    def supports_ignored_updates(self) -> bool:
        return self.installed

    def is_update_ignored(self) -> bool:
        return self.updates_ignored

    def get_update_ignore_key(self) -> str:
        return '{}:{}:{}'.format(self.installation, self.id, self.branch)

    def __eq__(self, other):
        if isinstance(other, FlatpakApplication):
            return self.id == other.id and self.installation == other.installation and self.branch == other.branch

    def get_disk_icon_path(self) -> str:
        if not self.runtime:
            return super(FlatpakApplication, self).get_disk_icon_path()

    def get_update_id(self, flatpak_version: Version) -> str:
        if flatpak_version >= VERSION_1_2:
            return f'{self.id}/{self.branch}/{self.installation}/{self.origin}'
        else:
            return f'{self.installation}/{self.ref}'

    def can_be_uninstalled(self) -> bool:
        return not self.update_component and super(FlatpakApplication, self).can_be_uninstalled()

    def update_ref(self):
        if self.id and self.arch and self.branch:
            self.ref = f'{self.id}/{self.arch}/{self.branch}'



import logging
from threading import Thread
from typing import List

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon, QPixmap, QCursor
from PyQt5.QtWidgets import QDialog, QLabel, QPushButton, QVBoxLayout, QProgressBar, QApplication, QWidget, \
    QSizePolicy, QHBoxLayout

from bauh.api.abstract.cache import MemoryCache
from bauh.api.http import HttpClient
from bauh.view.qt import qt_utils
from bauh.view.qt.components import new_spacer
from bauh.view.qt.thread import AnimateProgress
from bauh.view.qt.view_model import PackageView
from bauh.view.util.translation import I18n


class ScreenshotsDialog(QDialog):

    def __init__(self, pkg: PackageView, http_client: HttpClient, icon_cache: MemoryCache, i18n: I18n, screenshots: List[QPixmap], logger: logging.Logger):
        super(ScreenshotsDialog, self).__init__()
        self.setWindowTitle(str(pkg))
        self.screenshots = screenshots
        self.logger = logger
        self.loaded_imgs = []
        self.download_threads = []
        self.i18n = i18n
        self.http_client = http_client
        self.progress_bar = QProgressBar()
        self.progress_bar.setObjectName('progress_screenshots')
        self.progress_bar.setCursor(QCursor(Qt.WaitCursor))
        self.progress_bar.setMaximumHeight(10 if QApplication.instance().style().objectName().lower() == 'windows' else 6)
        self.progress_bar.setTextVisible(False)
        self.thread_progress = AnimateProgress()
        self.thread_progress.signal_change.connect(self._update_progress)
        self.thread_progress.start()

        # THERE ARE CRASHES WITH SOME RARE ICONS ( like insomnia ). IT CAN BE A QT BUG. IN THE MEANTIME, ONLY THE TYPE ICON WILL BE RENDERED
        #
        # icon_data = icon_cache.get(pkg.model.icon_url)
        #
        # if icon_data and icon_data.get('icon'):
        #     self.setWindowIcon(icon_data.get('icon'))
        # else:
        #     self.setWindowIcon(QIcon(pkg.model.get_type_icon_path()))
        self.setWindowIcon(QIcon(pkg.model.get_type_icon_path()))
        self.setLayout(QVBoxLayout())

        self.bt_close = QPushButton(self.i18n['screenshots.bt_close'])
        self.bt_close.setObjectName('close')
        self.bt_close.clicked.connect(self.close)
        self.bt_close.setCursor(QCursor(Qt.PointingHandCursor))

        self.upper_buttons = QWidget()
        self.upper_buttons.setObjectName('upper_buttons')
        self.upper_buttons.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.upper_buttons.setContentsMargins(0, 0, 0, 0)
        self.upper_buttons.setLayout(QHBoxLayout())
        self.upper_buttons.layout().setAlignment(Qt.AlignRight)
        self.upper_buttons.layout().addWidget(self.bt_close)
        self.layout().addWidget(self.upper_buttons)

        self.layout().addWidget(new_spacer())

        self.img = QLabel()
        self.img.setObjectName('image')
        self.layout().addWidget(self.img)
        self.layout().addWidget(new_spacer())

        self.container_buttons = QWidget()
        self.container_buttons.setObjectName('buttons_container')
        self.container_buttons.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.container_buttons.setContentsMargins(0, 0, 0, 0)
        self.container_buttons.setLayout(QHBoxLayout())

        self.bt_back = QPushButton(' < ' + self.i18n['screenshots.bt_back.label'].capitalize())
        self.bt_back.setObjectName('back')
        self.bt_back.setProperty('control', 'true')
        self.bt_back.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_back.clicked.connect(self.back)
        self.container_buttons.layout().addWidget(self.bt_back)
        self.container_buttons.layout().addWidget(new_spacer())

        self.img_label = QLabel()
        self.img_label.setObjectName("image_label")
        self.container_buttons.layout().addWidget(self.img_label)

        self.container_buttons.layout().addWidget(self.progress_bar)
        self.container_buttons.layout().addWidget(new_spacer())

        self.bt_next = QPushButton(self.i18n['screenshots.bt_next.label'].capitalize() + ' > ')
        self.bt_next.setObjectName('next')
        self.bt_next.setProperty('control', 'true')
        self.bt_next.setCursor(QCursor(Qt.PointingHandCursor))
        self.bt_next.clicked.connect(self.next)
        self.container_buttons.layout().addWidget(self.bt_next)

        self.layout().addWidget(self.container_buttons)

        self.img_idx = 0
        self.max_img_width = 800
        self.max_img_height = 600

        for idx, s in enumerate(self.screenshots):
            t = Thread(target=self._download_img, args=(idx, s), daemon=True)
            t.start()

        self.resize(self.max_img_width + 5, self.max_img_height + 5)
        self._load_img()
        qt_utils.centralize(self)

    def _update_progress(self, val: int):
        self.progress_bar.setValue(val)

    def _load_img(self):
        if len(self.loaded_imgs) > self.img_idx:
            img = self.loaded_imgs[self.img_idx]

            if isinstance(img, QPixmap):
                self.img_label.setText(f'{self.img_idx + 1}/{len(self.screenshots)}')
                self.img.setText('')
                self.img.setPixmap(img)
            else:
                self.img.setText(img)
                self.img.setPixmap(QPixmap())

            self.img.unsetCursor()
            self.thread_progress.stop = True
            self.progress_bar.setVisible(False)
        else:
            self.img.setPixmap(QPixmap())
            self.img.setCursor(QCursor(Qt.WaitCursor))
            self.img.setText('{} {}/{}...'.format(self.i18n['screenshots.image.loading'], self.img_idx + 1, len(self.screenshots)))
            self.progress_bar.setVisible(True)
            self.thread_progress.start()

        if len(self.screenshots) == 1:
            self.bt_back.setVisible(False)
            self.bt_next.setVisible(False)
        else:
            self.bt_back.setEnabled(self.img_idx != 0)
            self.bt_next.setEnabled(self.img_idx != len(self.screenshots) - 1)

    def _download_img(self, idx: int, url: str):
        self.logger.info('Downloading image [{}] from {}'.format(idx, url))
        res = self.http_client.get(url)

        if res:
            if not res.content:
                self.logger.warning('Image [{}] from {} has no content'.format(idx, url))
                self.loaded_imgs.append(self.i18n['screenshots.download.no_content'])
                self._load_img()
            else:
                self.logger.info('Image [{}] successfully downloaded'.format(idx))
                pixmap = QPixmap()
                pixmap.loadFromData(res.content)

                if pixmap.size().height() > self.max_img_height or pixmap.size().width() > self.max_img_width:
                    pixmap = pixmap.scaled(self.max_img_width, self.max_img_height, Qt.KeepAspectRatio, Qt.SmoothTransformation)

                self.loaded_imgs.append(pixmap)

                if self.img_idx == idx:
                    self._load_img()
        else:
            self.logger.info("Could not retrieve image [{}] from {}".format(idx, url))
            self.loaded_imgs.append(self.i18n['screenshots.download.no_response'])
            self._load_img()

    def back(self):
        self.img_idx -= 1
        self._load_img()

    def next(self):
        self.img_idx += 1
        self._load_img()



import socket
import traceback
from logging import Logger
from typing import Optional, List

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.connection import HTTPConnection
from urllib3.connectionpool import HTTPConnectionPool

from bauh.commons.system import run_cmd

URL_BASE = 'http://snapd/v2'


class SnapdConnection(HTTPConnection):
    def __init__(self):
        super(SnapdConnection, self).__init__('localhost')

    def connect(self):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self.sock.connect("/run/snapd.socket")


class SnapdConnectionPool(HTTPConnectionPool):
    def __init__(self):
        super(SnapdConnectionPool, self).__init__('localhost')

    def _new_conn(self):
        return SnapdConnection()


class SnapdAdapter(HTTPAdapter):

    def get_connection(self, url, proxies=None):
        return SnapdConnectionPool()


class SnapdClient:

    def __init__(self, logger: Logger):
        self.logger = logger
        self.session = self._new_session()

    def _new_session(self) -> Optional[Session]:
        try:
            session = Session()
            session.mount("http://snapd/", SnapdAdapter())
            return session
        except:
            self.logger.error("Could not establish a connection to 'snapd.socker'")
            traceback.print_exc()

    def query(self, query: str) -> Optional[List[dict]]:
        final_query = query.strip()

        if final_query and self.session:
            res = self.session.get(url=f'{URL_BASE}/find', params={'q': final_query})

            if res.status_code == 200:
                json_res = res.json()

                if json_res['status-code'] == 200:
                    return json_res['result']

    def find_by_name(self, name: str) -> Optional[List[dict]]:
        if name and self.session:
            res = self.session.get(f'{URL_BASE}/find?name={name}')

            if res.status_code == 200:
                json_res = res.json()

                if json_res['status-code'] == 200:
                    return json_res['result']

    def list_all_snaps(self) -> List[dict]:
        if self.session:
            res = self.session.get(f'{URL_BASE}/snaps')

            if res.status_code == 200:
                json_res = res.json()

                if json_res['status-code'] == 200:
                    return json_res['result']

        return []

    def list_only_apps(self) -> List[dict]:
        if self.session:
            res = self.session.get(f'{URL_BASE}/apps')

            if res.status_code == 200:
                json_res = res.json()

                if json_res['status-code'] == 200:
                    return json_res['result']
        return []

    def list_commands(self, name: str) -> List[dict]:
        if self.session:
            res = self.session.get(f'{URL_BASE}/apps?names={name}')

            if res.status_code == 200:
                json_res = res.json()

                if json_res['status-code'] == 200:
                    return [r for r in json_res['result'] if r['snap'] == name]
        return []


def is_running() -> bool:
    status = run_cmd('systemctl is-active snapd.service snapd.socket', print_error=False)
    if not status:
        return False
    else:
        for status in status.split('\n'):
            if status.strip().lower() == 'active':
                return True

        return False



import logging
from argparse import Namespace
from typing import Tuple

from PyQt5.QtWidgets import QApplication, QWidget

from bauh import ROOT_DIR, __app_name__, __version__
from bauh.api import user
from bauh.api.abstract.context import ApplicationContext
from bauh.api.http import HttpClient
from bauh.commons.internet import InternetChecker
from bauh.context import generate_i18n, DEFAULT_I18N_KEY, new_qt_application
from bauh.view.core import gems
from bauh.view.core.controller import GenericSoftwareManager
from bauh.view.core.downloader import AdaptableFileDownloader
from bauh.view.qt.prepare import PreparePanel
from bauh.view.qt.settings import SettingsWindow
from bauh.view.qt.window import ManageWindow
from bauh.view.util import resource, util
from bauh.view.util.cache import CacheCleaner, DefaultMemoryCacheFactory
from bauh.view.util.disk import DefaultDiskCacheLoaderFactory


def new_manage_panel(app_args: Namespace, app_config: dict, logger: logging.Logger) -> Tuple[QApplication, QWidget]:
    i18n = generate_i18n(app_config, resource.get_path('locale'))

    cache_cleaner = CacheCleaner()

    cache_factory = DefaultMemoryCacheFactory(expiration_time=int(app_config['memory_cache']['data_expiration']), cleaner=cache_cleaner)
    icon_cache = cache_factory.new(int(app_config['memory_cache']['icon_expiration']))

    http_client = HttpClient(logger)

    context = ApplicationContext(i18n=i18n,
                                 http_client=http_client,
                                 download_icons=bool(app_config['download']['icons']),
                                 app_root_dir=ROOT_DIR,
                                 cache_factory=cache_factory,
                                 disk_loader_factory=DefaultDiskCacheLoaderFactory(logger),
                                 logger=logger,
                                 distro=util.get_distro(),
                                 file_downloader=AdaptableFileDownloader(logger, bool(app_config['download']['multithreaded']),
                                                                         i18n, http_client, app_config['download']['multithreaded_client']),
                                 app_name=__app_name__,
                                 app_version=__version__,
                                 internet_checker=InternetChecker(offline=app_args.offline),
                                 root_user=user.is_root())

    managers = gems.load_managers(context=context, locale=i18n.current_key, config=app_config,
                                  default_locale=DEFAULT_I18N_KEY, logger=logger)

    if app_args.reset:
        util.clean_app_files(managers)
        exit(0)

    manager = GenericSoftwareManager(managers, context=context, config=app_config)

    app = new_qt_application(app_config=app_config, logger=logger, quit_on_last_closed=True)

    screen_size = app.primaryScreen().size()
    context.screen_width, context.screen_height = screen_size.width(), screen_size.height()

    if app_args.settings:  # only settings window
        manager.cache_available_managers()
        return app, SettingsWindow(manager=manager, i18n=i18n, window=None)
    else:
        manage_window = ManageWindow(i18n=i18n,
                                     manager=manager,
                                     icon_cache=icon_cache,
                                     screen_size=screen_size,
                                     config=app_config,
                                     context=context,
                                     http_client=http_client,
                                     icon=util.get_default_icon()[1],
                                     logger=logger)

        prepare = PreparePanel(screen_size=screen_size,
                               context=context,
                               manager=manager,
                               i18n=i18n,
                               manage_window=manage_window,
                               app_config=app_config)
        cache_cleaner.start()

        return app, prepare



import logging
from abc import ABC
from datetime import datetime
from logging import Logger
from typing import Optional, Union


class NullLoggerFactory(ABC):

    __instance: Optional[Logger] = None

    @classmethod
    def logger(cls) -> Logger:
        if cls.__instance is None:
            cls.__instance = logging.getLogger('__null__')
            cls.__instance.addHandler(logging.NullHandler())

        return cls.__instance


def deep_update(source: dict, overrides: dict):
    for key, value in overrides.items():
        if isinstance(value, dict):
            returned = deep_update(source.get(key, {}), value)
            source[key] = returned
        else:
            source[key] = overrides[key]
    return source


def size_to_byte(size: Union[float, int, str], unit: str, logger: Optional[Logger] = None) -> Optional[float]:
    lower_unit = unit.strip().lower()

    if isinstance(size, str):
        try:
            final_size = float(size.strip().replace(',', '.').replace(' ', ''))
        except ValueError:
            if logger:
                logger.error(f"Could not parse string size {size} to bytes")
            return
    else:
        final_size = float(size)

    if unit == 'b':
        return final_size / 8

    if unit == 'B':
        return final_size

    base = 1024 if lower_unit.endswith('ib') else 1000

    if lower_unit[0] == 'k':
        return final_size * base
    elif lower_unit[0] == 'm':
        return final_size * (base ** 2)
    elif lower_unit[0] == 'g':
        return final_size * (base ** 3)
    elif lower_unit[0] == 't':
        return final_size * (base ** 4)
    else:
        return final_size * (base ** 5)


def datetime_as_milis(date: datetime = datetime.utcnow()) -> int:
    return int(round(date.timestamp() * 1000))


def map_timestamp_file(file_path: str) -> str:
    path_split = file_path.split('/')
    return '/'.join(path_split[0:-1]) + '/' + path_split[-1].split('.')[0] + '.ts'



class Singleton(type):

    __instance = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.__instance:
            cls.__instance[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls.__instance[cls]



import glob
import os
import re
from typing import Optional, Dict, Tuple, Set

from bauh.api.paths import USER_THEMES_DIR
from bauh.view.util import resource
from bauh.view.util.translation import I18n

# RE_WIDTH_PERCENT = re.compile(r'[\d\\.]+%w') TODO percentage measures disabled for the moment (requires more testing)
# RE_HEIGHT_PERCENT = re.compile(r'[\d\\.]+%h') TODO percentage measures disabled for the moment (requires more testing)
RE_META_I18N_FIELDS = re.compile(r'((name|description)(\[\w+])?)')
RE_VAR_PATTERN = re.compile(r'^@[\w.\-_]+')
RE_QSS_EXT = re.compile(r'\.qss$')


class ThemeMetadata:

    def __init__(self, file_path: str, default: bool, default_name: Optional[str] = None,
                 default_description: Optional[str] = None, version: Optional[str] = None,
                 root_theme: Optional[str] = None, abstract: bool = False):
        self.names = {}
        self.default_name = default_name
        self.descriptions = {}
        self.default_description = default_description
        self.root_theme = root_theme
        self.version = version
        self.file_path = file_path
        self.file_dir = '/'.join(file_path.split('/')[0:-1])
        self.default = default
        self.key = self.file_path.split('/')[-1].split('.')[0] if self.default else self.file_path
        self.abstract = abstract

    def __eq__(self, other) -> bool:
        if isinstance(other, ThemeMetadata):
            return self.file_path == other.file_path

        return False

    def __hash__(self):
        return self.file_path.__hash__()

    def __repr__(self):
        return self.file_path if self.file_path else ''

    def get_i18n_name(self, i18n: I18n) -> str:
        if self.names:
            name = self.names.get(i18n.current_key, self.names.get(i18n.default_key))

            if name:
                return name

        if self.default_name:
            return self.default_name
        else:
            return self.file_path.split('/')[-1]

    def get_i18n_description(self, i18n: I18n) -> Optional[str]:
        if self.descriptions:
            des = self.descriptions.get(i18n.current_key, self.descriptions.get(i18n.default_key))

            if des:
                return des

        return self.default_description


def read_theme_metada(key: str, file_path: str) -> ThemeMetadata:
    meta_file = RE_QSS_EXT.sub('.meta', file_path)
    meta_obj = ThemeMetadata(file_path=file_path, default_name=key, default=not key.startswith('/'))

    if os.path.exists(meta_file):
        meta_dict = {}
        with open(meta_file) as f:
            for line in f.readlines():
                if line:
                    field_split = line.split('=')

                    if len(field_split) > 1:
                        meta_dict[field_split[0].strip()] = field_split[1].strip()

            if meta_dict:
                for field, val in meta_dict.items():
                    if field == 'version':
                        meta_obj.version = val
                    elif field == 'root_theme':
                        meta_obj.root_theme = val
                    elif field == 'name':
                        meta_obj.default_name = val
                    elif field == 'description':
                        meta_obj.default_description = val
                    elif field == 'abstract':
                        boolean = val.lower()

                        if boolean == 'true':
                            meta_obj.abstract = True
                        elif boolean == 'false':
                            meta_obj.abstract = False

                    else:
                        i18n_field = RE_META_I18N_FIELDS.findall(field)

                        if i18n_field:
                            if i18n_field[0][1] == 'name':
                                meta_obj.names[i18n_field[0][2][1:-1]] = val
                            else:
                                meta_obj.descriptions[i18n_field[0][2][1:-1]] = val

    return meta_obj


def read_default_themes() -> Dict[str, str]:
    return {f.split('/')[-1].split('.')[0].lower(): f for f in glob.glob(resource.get_path('style/**/*.qss'))}


def read_user_themes() -> Dict[str, str]:
    return {f: f for f in glob.glob('{}/**/*.qss'.format(USER_THEMES_DIR), recursive=True)}


def read_all_themes_metadata() -> Set[ThemeMetadata]:
    themes = set()

    for key, file_path in read_default_themes().items():
        themes.add(read_theme_metada(key=key, file_path=file_path))

    for key, file_path in read_user_themes().items():
        themes.add(read_theme_metada(key=key, file_path=file_path))

    return themes


def process_theme(file_path: str, theme_str: str, metadata: ThemeMetadata,
                  available_themes: Optional[Dict[str, str]]) -> Optional[Tuple[str, ThemeMetadata]]:
    if theme_str and metadata:
        root_theme = None
        if metadata.root_theme and metadata.root_theme in available_themes:
            root_file = available_themes[metadata.root_theme]

            if os.path.isfile(root_file):
                with open(root_file) as f:
                    root_theme_str = f.read()

                if root_theme_str:
                    root_metadata = read_theme_metada(key=metadata.root_theme, file_path=root_file)
                    root_theme = process_theme(file_path=root_file,
                                               theme_str=root_theme_str,
                                               metadata=root_metadata,
                                               available_themes=available_themes)

        var_map = _read_var_file(file_path)
        var_map['images'] = resource.get_path('img')
        var_map['style_dir'] = metadata.file_dir

        if var_map:
            var_list = [*var_map.keys()]
            var_list.sort(key=_by_str_len, reverse=True)

            for var in var_list:
                theme_str = theme_str.replace('@' + var, var_map[var])

        # TODO percentage measures disabled for the moment (requires more testing)
        # screen_size = QApplication.primaryScreen().size()
        # theme_str = process_width_percent_measures(theme_str, screen_size.width())
        # theme_str = process_height_percent_measures(theme_str, screen_size.height())

        return theme_str if not root_theme else '{}\n{}'.format(root_theme[0], theme_str), metadata


def _by_str_len(string: str) -> int:
    return len(string)


def _read_var_file(theme_file: str) -> dict:
    vars_file = theme_file.replace('.qss', '.vars')
    var_map = {}

    if os.path.isfile(vars_file):
        with open(vars_file) as f:
            for line in f.readlines():
                if line:
                    line_strip = line.strip()
                    if line_strip:
                        var_value = line_strip.split('=')

                        if var_value and len(var_value) == 2:
                            var, value = var_value[0].strip(), var_value[1].strip()

                            if var and value:
                                var_map[var] = value

    if var_map:
        process_var_of_vars(var_map)  # mapping keys that point to others

    return var_map


def process_var_of_vars(var_map: dict):
    while True:
        pending_vars, invalid = {}, set()

        for k, v in var_map.items():
            var_match = RE_VAR_PATTERN.match(v)

            if var_match:
                var_name = var_match.group()[1:]
                if var_name not in var_map or var_name == k:
                    invalid.add(k)
                else:
                    pending_vars[k] = var_name

        for key in invalid:
            del var_map[key]

        if not pending_vars:
            break

        resolved = 0

        for key, val in pending_vars.items():
            real_val = var_map[val]

            if not RE_VAR_PATTERN.match(real_val):
                var_map[key] = real_val
                resolved += 1

        if resolved == len(pending_vars):
            break


# TODO percentage measures disabled for the moment (requires more testing)
# def process_width_percent_measures(theme: str, screen_width: int) -> str:
#     width_measures = RE_WIDTH_PERCENT.findall(theme)
#
#     final_theme = theme
#     if width_measures:
#         for m in width_measures:
#             try:
#                 percent = float(m.split('%')[0])
#                 final_theme = final_theme.replace(m, '{}px'.format(round(screen_width * percent)))
#             except ValueError:
#                 traceback.print_exc()
#
#     return final_theme


# def process_height_percent_measures(theme: str, screen_height: int) -> str:
#     width_measures = RE_HEIGHT_PERCENT.findall(theme)
#
#     final_sheet = theme
#     if width_measures:
#         for m in width_measures:
#             try:
#                 percent = float(m.split('%')[0])
#                 final_sheet = final_sheet.replace(m, '{}px'.format(round(screen_height * percent)))
#             except ValueError:
#                 traceback.print_exc()
#
#     return final_sheet



import os
import re

RE_DESKTOP_EXEC = re.compile(r'(Exec\s*=(.+)(\n?))')
RE_MANY_SPACES = re.compile(r'\s+')


def find_appimage_file(folder: str) -> str:
    for r, d, files in os.walk(folder):
        for f in files:
            if f.lower().endswith('.appimage'):
                return '{}/{}'.format(folder, f)


def replace_desktop_entry_exec_command(desktop_entry: str, appname: str, file_path: str) -> str:
    execs = RE_DESKTOP_EXEC.findall(desktop_entry)

    if not execs:
        return desktop_entry

    final_entry = desktop_entry
    treated_name = appname.strip().lower()

    for exec_groups in execs:
        full_match = exec_groups[0]
        cmd = RE_MANY_SPACES.sub(' ', exec_groups[1].strip())
        if cmd:
            words = cmd.split(' ')
            changed = False

            for idx in range(len(words)):
                if words[idx].lower() == treated_name:
                    words[idx] = '"{}"'.format(file_path)
                    changed = True
                    break

            if not changed:
                words = ['"{}"'.format(file_path)]

            final_entry = final_entry.replace(full_match, full_match.replace(exec_groups[1], ' '.join(words)))

    return final_entry



from unittest import TestCase

from bauh.commons.util import size_to_byte


class SizeToByteTest(TestCase):

    def test_must_return_right_number_of_bytes_for_bit_size(self):
        self.assertEqual(0.125, size_to_byte(1, 'b'))
        self.assertEqual(0.250, size_to_byte(2, 'b'))
        self.assertEqual(0.40625, size_to_byte(3.25, 'b'))

    def test_must_return_right_number_of_bytes_for_byte_based_units(self):
        self.assertEqual(1.0, size_to_byte(1, 'B'))
        self.assertEqual(587, size_to_byte(587, 'B'))
        self.assertEqual(1000, size_to_byte(1, 'K'))
        self.assertEqual(57300, size_to_byte(57.3, 'K'))
        self.assertEqual(1000000, size_to_byte(1, 'M'))
        self.assertEqual(57300000, size_to_byte(57.3, 'M'))
        self.assertEqual(1000000000, size_to_byte(1, 'G'))
        self.assertEqual(57300000000, size_to_byte(57.3, 'G'))
        self.assertEqual(1000000000000, size_to_byte(1, 'T'))
        self.assertEqual(57300000000000, size_to_byte(57.3, 'T'))
        self.assertEqual(1000000000000000, size_to_byte(1, 'P'))
        self.assertEqual(5670000000000000, size_to_byte(5.67, 'P'))

    def test_must_return_right_number_of_bytes_for_bibyte_units(self):
        self.assertEqual(1024.0, size_to_byte(1, 'KiB'))
        self.assertEqual(1579683.84, size_to_byte(1542.66, 'KiB'))
        self.assertEqual(1048576, size_to_byte(1, 'MiB'))
        self.assertEqual(4372561.92, size_to_byte(4.17, 'MiB'))
        self.assertEqual(1073741824, size_to_byte(1, 'GiB'))
        self.assertEqual(2168958484.48, size_to_byte(2.02, 'GiB'))
        self.assertEqual(1099511627776, size_to_byte(1, 'TiB'))
        self.assertEqual(3375500697272.32, size_to_byte(3.07, 'TiB'))
        self.assertEqual(1125899906842624, size_to_byte(1, 'PiB'))
        self.assertEqual(1294784892869017.5, size_to_byte(1.15, 'PiB'))

    def test_must_return_converted_string_sizes(self):
        self.assertEqual(1000, size_to_byte('1', 'K'))
        self.assertEqual(57300, size_to_byte('57.3', 'K'))
        self.assertEqual(57300000, size_to_byte('57.3', 'M'))
        self.assertEqual(57300000000, size_to_byte('57.3', 'G'))
        self.assertEqual(57300000000000, size_to_byte('57.3', 'T'))
        self.assertEqual(5670000000000000, size_to_byte('5.67', 'P'))

    def test_must_treat_string_sizes_before_converting(self):
        self.assertEqual(57300, size_to_byte(' 57 , 3  ', ' K '))



from unittest import TestCase

from bauh.gems.arch import sorting


class SortingTest(TestCase):

    def test_sort__aur_not_related(self):
        pkgs = {'google-chrome': {'d': {'alsa-lib', 'gtk3', 'libcups'}, 'p': {'google-chrome': 'google-chrome'}, 'r': 'extra'},
                'git-cola': {'d': {'git', 'python-pyqt5', 'icu qt5-svg'}, 'p': {'git-cola': 'git-cola'}, 'r': 'extra'},
                'kazam': {'d': {'python', 'python-cairo'}, 'p': {'kazam': 'kazam'}, 'r': 'extra'}}

        sorted_list = sorting.sort(pkgs.keys(), pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        for pkg in sorted_list:
            self.assertIn(pkg[0], pkgs)

    def test_sort__all_packages_no_deps(self):
        pkgs = {'xpto': {'d': set(), 'p': {'xpto': 'xpto'}, 'r': 'extra'},
                'abc': {'d': None, 'p': {'abc': 'abc'}, 'r': 'extra'}}

        sorted_list = sorting.sort(pkgs.keys(), pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        for pkg in sorted_list:
            self.assertIn(pkg[0], pkgs)

    def test_sort__one_of_three_related(self):
        pkgs = {'def': {'d': {'jkl'}, 'p': {'def': 'def'}, 'r': 'extra'},
                'abc': {'d': {'ghi', 'xpto'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': None, 'p': {'ghi': 'ghi'}, 'r': 'extra'}}

        for _ in range(5):  # testing n times to see if the same result is produced
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            ghi = [p for p in sorted_list if p[0] == 'ghi']
            self.assertEqual(1, len(ghi))

            ghi_idx = sorted_list.index(ghi[0])

            abc = [p for p in sorted_list if p[0] == 'abc']
            self.assertEqual(1, len(abc))

            abc_idx = sorted_list.index(abc[0])
            self.assertGreater(abc_idx, ghi_idx)

    def test_sort__two_of_three_related(self):
        """
            dep order = abc -> ghi -> def
            expected: def, ghi, abc
        """
        pkgs = {'def': {'d': {'jkl'}, 'p': {'def': 'def'}, 'r': 'extra'},
                'abc': {'d': {'ghi', 'xpto'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': {'def'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'}}

        for _ in range(5):  # testing n times to see if the same result is produced
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            self.assertEqual(sorted_list[0][0], 'def')
            self.assertEqual(sorted_list[1][0], 'ghi')
            self.assertEqual(sorted_list[2][0], 'abc')

    def test_sort__all_related(self):
        pkgs = {'lutris': {'d': {'unzip', 'python-requests'}, 'p': {'lutris': 'lutris'}, 'r': 'community'},
                'python-requests': {'d': {'python-urllib3', 'python-chardet'},
                                    'p': {'python-requests': 'python-requests'}, 'r': 'extra'},
                'python-chardet': {'d': {'python-setuptools'}, 'p': {'python-chardet': 'python-chardet'},
                                   'r': 'extra'}}

        sorted_list = sorting.sort(pkgs.keys(), pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        for pkg in sorted_list:
            self.assertIn(pkg[0], pkgs)

        self.assertEqual('python-chardet', sorted_list[0][0])
        self.assertEqual('python-requests', sorted_list[1][0])
        self.assertEqual('lutris', sorted_list[2][0])

    def test_sorting__pkg_should_be_after_its_latest_dependency(self):
        pkgs = {'abc': {'d': {'def', 'ghi'}, 'p': {'abc': 'abc'}, 'r': 'community'},
                'ghi': {'d': {'xpto'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'},
                'xpto': {'d': {'zzz'}, 'p': {'xpto': 'xpto'}, 'r': 'extra'},
                'def': {'d': None, 'p': {'def': 'def'}, 'r': 'extra'}}

        for _ in range(5):  # to ensure the result is always the same
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            self.assertEqual('def', sorted_list[0][0])
            self.assertEqual('xpto', sorted_list[1][0])
            self.assertEqual('ghi', sorted_list[2][0])
            self.assertEqual('abc', sorted_list[3][0])

    def test_sort__two_relying_on_the_same_package(self):
        """
            dep order:
                abc -> ghi
                jkl -> ghi
                ghi -> def
                def -> mno
            expected: def, ghi, (abc | jkl )
        """
        pkgs = {'def': {'d': {'mno'}, 'p': {'def': 'def'}, 'r': 'extra'},
                'abc': {'d': {'ghi', 'xpto'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': {'def'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'},
                'jkl': {'d': {'ghi'}, 'p': {'jkl': 'jkl'}, 'r': 'extra'}}

        for _ in range(5):  # testing n times to see if the same result is produced
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            self.assertEqual(sorted_list[0][0], 'def')
            self.assertEqual(sorted_list[1][0], 'ghi')

            self.assertNotEqual(sorted_list[2][0],  sorted_list[3][0])
            self.assertIn(sorted_list[2][0], {'abc', 'jkl'})
            self.assertIn(sorted_list[3][0], {'abc', 'jkl'})

    def test_sort__with_cycle(self):
        """
            dep order:
                abc -> def -> ghi -> jkl -> abc
        """
        pkgs = {'def': {'d': {'ghi'}, 'p': {'def': 'def'}, 'r': 'extra'},
                'abc': {'d': {'def'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': {'jkl'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'},
                'jkl': {'d': {'abc'}, 'p': {'jkl': 'jkl'}, 'r': 'extra'}}

        sorted_list = sorting.sort(pkgs.keys(), pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        for pkg in sorted_list:
            self.assertIn(pkg[0], pkgs)

    def test_sort__dep_provided_as_a_different_name(self):
        """
            dep order:
                abc -> fed
                def (fed)
                ghi -> abc
            expected: def, abc, ghi
        """
        pkgs = {'def': {'d': None, 'p': {'def': 'def', 'fed': 'def'}, 'r': 'extra'},
                'abc': {'d': {'fed'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': {'abc'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'}}

        for _ in range(5):
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            self.assertEqual(sorted_list[0][0], 'def')
            self.assertEqual(sorted_list[1][0], 'abc')
            self.assertEqual(sorted_list[2][0], 'ghi')
            
    def test_sort__aur_pkgs_should_be_always_in_the_end(self):
        """
            dep order:
                abc -> fed
                def (fed)
                ghi -> abc
            expected: def, abc, ghi
        """
        pkgs = {'def': {'d': None, 'p': {'def': 'def'}, 'r': 'aur'},
                'abc': {'d': {'ghi'}, 'p': {'abc': 'abc'}, 'r': 'extra'},
                'ghi': {'d': {'xxx'}, 'p': {'ghi': 'ghi'}, 'r': 'extra'}}

        for _ in range(5):
            sorted_list = sorting.sort(pkgs.keys(), pkgs)
            self.assertIsInstance(sorted_list, list)
            self.assertEqual(len(pkgs), len(sorted_list))

            for pkg in sorted_list:
                self.assertIn(pkg[0], pkgs)

            self.assertEqual(sorted_list[0][0], 'ghi')
            self.assertEqual(sorted_list[1][0], 'abc')
            self.assertEqual(sorted_list[2][0], 'def')



from bauh.commons.config import YAMLConfigManager
from bauh.gems.web import CONFIG_FILE


class WebConfigManager(YAMLConfigManager):

    def __init__(self):
        super(WebConfigManager, self).__init__(config_file_path=CONFIG_FILE)

    def get_default_config(self) -> dict:
        return {
            'environment': {
                'system': False,
                'electron': {'version': None},
                'cache_exp': 24
            },
            'suggestions': {
                'cache_exp': 24
            }
        }



import re

HTML_RE = re.compile(r'<[^>]+>')


def strip_html(string: str):
    return HTML_RE.sub('', string)


def bold(text: str) -> str:
    return '<span style="font-weight: bold">{}</span>'.format(text)


def link(url: str) -> str:
    return '<a href="{}">{}</a>'.format(url, url)



from unittest import TestCase
from unittest.mock import patch, Mock

from bauh import __app_name__
from bauh.gems.flatpak import flatpak, VERSION_1_2


class FlatpakTest(TestCase):

    @patch(f'{__app_name__}.gems.flatpak.flatpak.SimpleProcess')
    @patch(f'{__app_name__}.gems.flatpak.flatpak.ProcessHandler.handle_simple', return_value=(True, """
    Looking for updates...
    
    \tID\tArch\tBranch\tRemote\tDownload
    1.\t \torg.xpto.Xnote\tx86_64\tstable\tflathub\t< 4.3 MB
    
    """))
    def test_map_update_download_size__for_flatpak_1_2(self, SimpleProcess: Mock, handle_simple: Mock):
        download_size = flatpak.map_update_download_size(app_ids={'org.xpto.Xnote'}, installation='user', version=VERSION_1_2)
        SimpleProcess.assert_called_once()
        handle_simple.assert_called_once()

        self.assertEqual({'org.xpto.Xnote': 4300000}, download_size)



from unittest import TestCase
from unittest.mock import Mock

from bauh.gems.flatpak.controller import FlatpakManager
from bauh.gems.flatpak.model import FlatpakApplication


class FlatpakManagerSortUpdateOrderTest(TestCase):

    def setUp(self):
        self.manager = FlatpakManager(Mock())

    def test__sort_deps__only_apps(self):
        pkgs = [
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            FlatpakApplication(id="com.spotify.Client", name='com.spotify.Client', runtime=False),
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        for pkg in sorted_list:
            self.assertIn(pkg, pkgs)

    def test__sort_deps__one_app_one_runtime(self):
        pkgs = [
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            FlatpakApplication(id="org.gnome.Platform", name='org.gnome.Platform', runtime=True),
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        self.assertEqual(pkgs[1], sorted_list[0])
        self.assertEqual(pkgs[0], sorted_list[1])

    def test__sort_deps__two_apps_one_runtime(self):
        pkgs = [
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            FlatpakApplication(id="org.gnome.Platform", name='org.gnome.Platform', runtime=True),
            FlatpakApplication(id="com.spotify.Client", name='com.spotify.Client', runtime=False)
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        self.assertEqual(pkgs[1], sorted_list[0])
        self.assertEqual(pkgs[0], sorted_list[1])
        self.assertEqual(pkgs[2], sorted_list[2])

    def test_sort_deps_two_apps_two_runtimes(self):
        pkgs = [
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            FlatpakApplication(id="org.gnome.Platform", name='Platform', runtime=True),
            FlatpakApplication(id="com.spotify.Client", name='Spotify', runtime=False),
            FlatpakApplication(id="org.freedesktop.Platform.GL.default", name='default', runtime=True)
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        self.assertEqual(pkgs[1], sorted_list[0])
        self.assertEqual(pkgs[3], sorted_list[1])
        self.assertEqual(pkgs[0], sorted_list[2])
        self.assertEqual(pkgs[2], sorted_list[3])

    def test_sort_deps_two_apps_two_runtimes_one_partial(self):
        gnome_platform = FlatpakApplication(id="org.gnome.Platform", name='Platform', runtime=True)
        gnome_locale = FlatpakApplication(id="org.gnome.Platform.Locale", name="Locale", runtime=True, partial=True)
        gnome_locale.base_id = gnome_platform.id

        pkgs = [
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            gnome_platform,
            FlatpakApplication(id="com.spotify.Client", name='Spotify', runtime=False),
            FlatpakApplication(id="org.freedesktop.Platform.GL.default", name='default', runtime=True),
            gnome_locale
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        self.assertEqual(pkgs[4], sorted_list[0])
        self.assertEqual(pkgs[1], sorted_list[1])
        self.assertEqual(pkgs[3], sorted_list[2])
        self.assertEqual(pkgs[0], sorted_list[3])
        self.assertEqual(pkgs[2], sorted_list[4])

    def test_sort_deps_two_apps_two_runtimes_two_partials(self):
        gnome_platform = FlatpakApplication(id="org.gnome.Platform", name='Platform', runtime=True)
        gnome_locale = FlatpakApplication(id="org.gnome.Platform.Locale", name="Locale", runtime=True, partial=True)
        gnome_locale.base_id = gnome_platform.id

        platform_default = FlatpakApplication(id="org.freedesktop.Platform.GL.default", name='default', runtime=True)
        platform_locale = FlatpakApplication(id="org.freedesktop.Platform.GL.Locale", name='Locale', runtime=True, partial=True)
        platform_locale.base_id = platform_default.id

        pkgs = [
            platform_locale,
            FlatpakApplication(id="org.gnome.gedit", name='Gedit', runtime=False),
            gnome_platform,
            FlatpakApplication(id="com.spotify.Client", name='Spotify', runtime=False),
            platform_default,
            gnome_locale
        ]

        sorted_list = self.manager.sort_update_order(pkgs)
        self.assertIsInstance(sorted_list, list)
        self.assertEqual(len(pkgs), len(sorted_list))

        self.assertEqual(gnome_locale.id, sorted_list[0].id)
        self.assertEqual(gnome_platform.id, sorted_list[1].id)
        self.assertEqual(platform_locale.id, sorted_list[2].id)
        self.assertEqual(platform_default.id, sorted_list[3].id)
        self.assertEqual('org.gnome.gedit', sorted_list[4].id)
        self.assertEqual('com.spotify.Client', sorted_list[5].id)



import re
from typing import Set

RE_PKGBUILD_OPTDEPS = re.compile(r"optdepends = (.+)")
RE_PKGBUILD_OPTDEPS_x86_64 = re.compile(r"optdepends_x86_64 = (.+)")
RE_PKGBUILD_OPTDEPS_i686 = re.compile(r"optdepends_i686 = (.+)")


def read_optdeps_as_dict(srcinfo: str, x86_64: bool) -> dict:
    res = {}
    for optdep in read_optdeps(srcinfo, x86_64):
        split_dep = optdep.split(':')
        res[split_dep[0].strip()] = split_dep[1].strip() if len(split_dep) > 1 else None

    return res


def read_optdeps(srcinfo: str, x86_64: bool) -> Set[str]:
    optdeps = set(RE_PKGBUILD_OPTDEPS.findall(srcinfo))

    if x86_64:
        optdeps.update(set(RE_PKGBUILD_OPTDEPS_x86_64.findall(srcinfo)))
    else:
        optdeps.update(set(RE_PKGBUILD_OPTDEPS_i686.findall(srcinfo)))

    return optdeps



import locale
from unittest import TestCase

from bauh.commons.view_utils import get_human_size_str


class GetHumanSizeStrTest(TestCase):

    def setUp(self):
        try:
            locale.setlocale(locale.LC_NUMERIC, None)
        except:
            print("Error: could not set locale.LC_NUMERIC to None")

    def test__must_properly_display_B(self):
        self.assertEqual('1 B', get_human_size_str(1))
        self.assertEqual('999 B', get_human_size_str(999))
        self.assertEqual('-999 B', get_human_size_str(-999))

    def test__must_properly_convert_to_kB(self):
        self.assertEqual('1.00 kB', get_human_size_str(1000))
        self.assertEqual('-1.00 kB', get_human_size_str(-1000))
        self.assertEqual('1.50 kB', get_human_size_str(1500))
        self.assertEqual('1.75 kB', get_human_size_str(1750))
        self.assertEqual('57.30 kB', get_human_size_str(57300))

    def test__must_properly_convert_to_MB(self):
        self.assertEqual('1.00 MB', get_human_size_str(1000 ** 2))
        self.assertEqual('-1.00 MB', get_human_size_str(1000 ** 2 * -1))
        self.assertEqual('1.20 MB', get_human_size_str(1000 ** 2 * 1.2))
        self.assertEqual('1.50 MB', get_human_size_str(1000 ** 2 * 1.5))
        self.assertEqual('57.30 MB', get_human_size_str(1000 ** 2 * 57.3))

    def test__must_properly_convert_to_GB(self):
        self.assertEqual('1.00 GB', get_human_size_str(1000 ** 3))
        self.assertEqual('-1.00 GB', get_human_size_str(1000 ** 3 * -1))
        self.assertEqual('57.30 GB', get_human_size_str(1000 ** 3 * 57.3))

    def test__must_properly_convert_to_TB(self):
        self.assertEqual('1.00 TB', get_human_size_str(1000 ** 4))
        self.assertEqual('-1.00 TB', get_human_size_str(1000 ** 4 * -1))
        self.assertEqual('57.30 TB', get_human_size_str(1000 ** 4 * 57.3))

    def test__must_properly_convert_to_PB(self):
        self.assertEqual('1.00 PB', get_human_size_str(1000 ** 5))
        self.assertEqual('-1.00 PB', get_human_size_str(1000 ** 5 * -1))
        self.assertEqual('57.30 PB', get_human_size_str(1000 ** 5 * 57.3))

    def test__must_concatenate_the_plus_sign_if_positive_sign_is_true_and_value_is_positive(self):
        self.assertEqual('+999 B', get_human_size_str(999, positive_sign=True))
        self.assertEqual('+1.00 kB', get_human_size_str(1000, positive_sign=True))

    def test__must_not_concatenate_the_plus_sign_if_positive_sign_is_true_and_value_is_negative(self):
        self.assertEqual('-999 B', get_human_size_str(-999, positive_sign=True))
        self.assertEqual('-1.00 kB', get_human_size_str(-1000, positive_sign=True))




import os

from bauh import __app_name__
from bauh.api.paths import DESKTOP_ENTRIES_DIR, CONFIG_DIR, TEMP_DIR, CACHE_DIR, SHARED_FILES_DIR
from bauh.commons import resource
from bauh.commons.util import map_timestamp_file

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
WEB_SHARED_DIR = f'{SHARED_FILES_DIR}/web'
WEB_CACHE_DIR = f'{CACHE_DIR}/web'
INSTALLED_PATH = f'{WEB_SHARED_DIR}/installed'
ENV_PATH = f'{WEB_SHARED_DIR}/env'
FIX_FILE_PATH = WEB_SHARED_DIR + '/fixes/{electron_branch}/{app_id}.js'
NODE_DIR_PATH = f'{ENV_PATH}/node'
NODE_PATHS = {f'{NODE_DIR_PATH}/bin'}
NODE_BIN_PATH = f'{NODE_DIR_PATH}/bin/node'
NPM_BIN_PATH = f'{NODE_DIR_PATH}/bin/npm'
NODE_MODULES_PATH = f'{ENV_PATH}/node_modules'
NATIVEFIER_BIN_PATH = f'{NODE_MODULES_PATH}/.bin/nativefier'
ELECTRON_CACHE_DIR = f'{ENV_PATH}/electron'
URL_ENVIRONMENT_SETTINGS = f'https://raw.githubusercontent.com/vinifmor/bauh-files/master/web/env/v2/environment.yml'
DESKTOP_ENTRY_PATH_PATTERN = f'{DESKTOP_ENTRIES_DIR}/{__app_name__}.web.' + '{name}.desktop'
URL_FIX_PATTERN = "https://raw.githubusercontent.com/vinifmor/bauh-files/master/web/env/v2/fix/{domain}/{electron_branch}/fix.js"
URL_PROPS_PATTERN = "https://raw.githubusercontent.com/vinifmor/bauh-files/master/web/env/v2/fix/{domain}/{electron_branch}/properties"
URL_SUGGESTIONS = "https://raw.githubusercontent.com/vinifmor/bauh-files/master/web/env/v2/suggestions.yml"
UA_CHROME = 'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/96.0.4664.110 Safari/537.36'
TEMP_PATH = f'{TEMP_DIR}/web'
SEARCH_INDEX_FILE = f'{WEB_CACHE_DIR}/index.yml'
SUGGESTIONS_CACHE_FILE = f'{WEB_CACHE_DIR}/suggestions.yml'
SUGGESTIONS_CACHE_TS_FILE = map_timestamp_file(SUGGESTIONS_CACHE_FILE)
CONFIG_FILE = f'{CONFIG_DIR}/web.yml'
ENVIRONMENT_SETTINGS_CACHED_FILE = f'{WEB_CACHE_DIR}/environment.yml'
ENVIRONMENT_SETTINGS_TS_FILE = f'{WEB_CACHE_DIR}/environment.ts'


def get_icon_path() -> str:
    return resource.get_path('img/web.svg', ROOT_DIR)



import json

from bauh.cli import __app_name__
from bauh.view.core.controller import GenericSoftwareManager


class CLIManager:

    def __init__(self, manager: GenericSoftwareManager):
        self.manager = manager

    def _print(self, msg: str):
        print('[{}] {}'.format(__app_name__, msg))

    def list_updates(self, output_format: str):
        updates = self.manager.list_updates()

        json_output = output_format == 'json'

        if not updates and not json_output:
            self._print('No updates available')
            return

        if not json_output:
            self._print('There are {} updates available:\n'.format(len(updates)))

            for idx, u in enumerate(updates):
                print('{}. Name: {}\tVersion: {}\tType: {}'.format(idx+1, u.name, u.version, u.type))
        else:
            print(json.dumps([u.__dict__ for u in updates]))



import os

from bauh.api.paths import CACHE_DIR, CONFIG_DIR
from bauh.commons import resource

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

DEBIAN_CACHE_DIR = f'{CACHE_DIR}/debian'
APP_INDEX_FILE = f'{DEBIAN_CACHE_DIR}/apps_idx.json'
CONFIG_FILE = f'{CONFIG_DIR}/debian.yml'
PACKAGE_SYNC_TIMESTAMP_FILE = f'{DEBIAN_CACHE_DIR}/sync_pkgs.ts'
DEBIAN_ICON_PATH = resource.get_path('img/debian.svg', ROOT_DIR)



import os.path
import shutil
import traceback
from operator import attrgetter
from pathlib import Path
from shutil import which
from subprocess import Popen
from threading import Thread
from typing import List, Optional, Tuple, Set, Type, Dict, Iterable, Generator

from bauh.api.abstract.context import ApplicationContext
from bauh.api.abstract.controller import SoftwareManager, SoftwareAction, TransactionResult, UpgradeRequirements, \
    SearchResult, UpgradeRequirement, SettingsView, SettingsController
from bauh.api.abstract.disk import DiskCacheLoader
from bauh.api.abstract.handler import TaskManager, ProcessWatcher
from bauh.api.abstract.model import SoftwarePackage, PackageSuggestion, PackageUpdate, PackageHistory, \
    CustomSoftwareAction
from bauh.api.abstract.view import TextInputComponent, PanelComponent, FormComponent, MessageType, \
    SingleSelectComponent, InputOption, SelectViewType
from bauh.api.paths import CONFIG_DIR
from bauh.commons.html import bold
from bauh.commons.system import ProcessHandler
from bauh.commons.util import NullLoggerFactory
from bauh.commons.view_utils import get_human_size_str
from bauh.gems.debian import DEBIAN_ICON_PATH
from bauh.gems.debian.aptitude import Aptitude, AptitudeOutputHandlerFactory, AptitudeAction
from bauh.gems.debian.common import fill_show_data
from bauh.gems.debian.config import DebianConfigManager
from bauh.gems.debian.gui import DebianViewBridge
from bauh.gems.debian.index import ApplicationIndexer, ApplicationIndexError, ApplicationsMapper
from bauh.gems.debian.model import DebianPackage, DebianApplication
from bauh.gems.debian.suggestions import DebianSuggestionsDownloader
from bauh.gems.debian.tasks import UpdateApplicationIndex, MapApplications, SynchronizePackages


class DebianPackageManager(SoftwareManager, SettingsController):

    def __init__(self, context: ApplicationContext):
        super(DebianPackageManager, self).__init__(context)
        self._i18n = context.i18n
        self._log = context.logger
        self._types: Optional[Set[Type[SoftwarePackage]]] = None
        self._enabled = True
        self._app_indexer: Optional[ApplicationIndexer] = None
        self._apps_index: Optional[Dict[str, DebianApplication]] = None
        self._configman: Optional[DebianConfigManager] = None
        self._action_launch_sources: Optional[CustomSoftwareAction] = None
        self._default_actions: Optional[Iterable[CustomSoftwareAction]] = None
        self._view: Optional[DebianViewBridge] = None
        self._app_mapper: Optional[ApplicationsMapper] = None
        self._aptitude: Optional[Aptitude] = None
        self._output_handler: Optional[AptitudeOutputHandlerFactory] = None
        self._known_sources_apps: Optional[Tuple[str, ...]] = None
        self._install_show_attrs: Optional[Set[str]] = None
        self._file_ignored_updates: Optional[str] = None
        self._suggestions_downloader: Optional[DebianSuggestionsDownloader] = None

    def _update_apps_index(self, apps: Iterable[DebianApplication]):
        self._apps_index = {app.name: app for app in apps} if apps else dict()

    def search(self, words: str, disk_loader: Optional[DiskCacheLoader], limit: int, is_url: bool) -> SearchResult:
        config_ = dict()
        fill_config = Thread(target=self._fill_config, args=(config_,))
        fill_config.start()

        res = SearchResult.empty()

        if not is_url:
            for pkg in self.aptitude.search(words):
                if fill_config.is_alive():
                    fill_config.join()
                    
                pkg.global_purge = bool(config_.get('remove.purge', False))
                if pkg.installed:
                    pkg.bind_app(self.apps_index.get(pkg.name))
                    res.installed.append(pkg)
                else:
                    res.new.append(pkg)

        return res

    def _fill_ignored_updates(self, output: Set[str]):

        try:
            with open(self.file_ignored_updates) as f:
                ignored_str = f.read()
        except FileNotFoundError:
            return

        if ignored_str:
            for line in ignored_str.split('\n'):
                line_clean = line.strip()

                if line_clean:
                    output.add(line_clean)

    def _fill_config(self, config_: dict):
        config_.update(self.configman.get_config())

    def read_installed(self, disk_loader: Optional[DiskCacheLoader], pkg_types: Optional[Set[Type[SoftwarePackage]]],
                       internet_available: bool, limit: int = -1, only_apps: bool = False,
                       names: Optional[Iterable[str]] = None) -> SearchResult:

        config_ = dict()
        fill_config = Thread(target=self._fill_config, args=(config_,))
        fill_config.start()

        ignored_updates = set()
        fill_ignored_updates = Thread(target=self._fill_ignored_updates, args=(ignored_updates,))
        fill_ignored_updates.start()

        threads = (fill_config, fill_ignored_updates)

        res = SearchResult(installed=[], new=None, total=0)

        for pkg in self.aptitude.read_installed():
            for t in threads:
                if t.is_alive():
                    t.join()

            pkg.bind_app(self.apps_index.get(pkg.name))
            pkg.global_purge = bool(config_.get('remove.purge', False))
            pkg.updates_ignored = bool(ignored_updates and pkg.name in ignored_updates)
            res.installed.append(pkg)

        return res

    def downgrade(self, pkg: SoftwarePackage, root_password: str, handler: ProcessWatcher) -> bool:
        return False

    def upgrade(self, requirements: UpgradeRequirements, root_password: str, watcher: ProcessWatcher) -> bool:
        handler = ProcessHandler(watcher)

        targets = (r.pkg.name for r in (*requirements.to_upgrade, *(requirements.to_install or ())))

        with self.output_handler.start(watcher=watcher, targets=targets, action=AptitudeAction.UPGRADE) as handle:
            to_upgrade = (r.pkg.name for r in requirements.to_upgrade)
            success, _ = handler.handle_simple(self.aptitude.upgrade(packages=to_upgrade,
                                                                     root_password=root_password),
                                               output_handler=handle)
        return success

    def _fill_updates(self, output: Dict[str, str]):
        for name, version in self.aptitude.read_updates():
            output[name] = version

    def uninstall(self, pkg: DebianPackage, root_password: str, watcher: ProcessWatcher,
                  disk_loader: Optional[DiskCacheLoader], purge: bool = False) -> TransactionResult:

        config_ = self.configman.get_config()
        purge_ = purge or config_.get('remove.purge', False)

        watcher.change_substatus(self._i18n['debian.simulate_operation'])

        transaction = self.aptitude.simulate_removal((pkg.name,), purge=purge_)

        if not transaction or not transaction.to_remove:
            return TransactionResult.fail()

        if pkg not in transaction.to_remove:
            watcher.show_message(title=self._i18n['popup.title.error'],
                                 body=self._i18n['debian.remove.impossible'].format(pkg=bold(pkg.name)),
                                 type_=MessageType.ERROR)
            return TransactionResult.fail()

        watcher.change_substatus('')

        deps = tuple(p for p in transaction.to_remove if p.name != pkg.name)

        if deps:
            # updates are required to be filled in case the dependencies are currently displayed on the view
            updates = dict()
            fill_updates = Thread(target=self._fill_updates, args=(updates,))
            fill_updates.start()

            deps_data = self.aptitude.show((p.name for p in deps), attrs=('description', 'maintainer', 'section'))

            if deps_data:
                for p in deps:
                    fill_show_data(p, deps_data.get(p.name))

            if not self.view.confirm_removal(source_pkg=pkg.name, dependencies=deps, watcher=watcher):
                return TransactionResult.fail()

            fill_updates.join()

            if updates:
                for p in deps:
                    latest_version = updates.get(p.name)

                    if latest_version is not None and p.version != latest_version:
                        p.latest_version = latest_version
                        p.update = True

        watcher.change_substatus(self._i18n['debian.uninstall.removing'])

        handler = ProcessHandler(watcher)
        to_remove = tuple(p.name for p in transaction.to_remove)
        with self.output_handler.start(watcher=watcher, targets=to_remove, action=AptitudeAction.REMOVE) as handle:
            removed, _ = handler.handle_simple(self.aptitude.remove(packages=to_remove, root_password=root_password,
                                                                    purge=purge_),
                                               output_handler=handle)

        if not removed:
            return TransactionResult.fail()

        watcher.change_substatus(self._i18n['debian.uninstall.validating'])

        current_installed_names = set(self.aptitude.read_installed_names())

        watcher.change_substatus('')

        all_removed, apps_removed, not_removed_names = [], set(), set()

        for p in transaction.to_remove:
            if p.name not in current_installed_names:
                instance = p if p != pkg else pkg

                all_removed.append(instance)

                if instance.app:
                    apps_removed.add(instance.app)

                instance.installed = False
                instance.version = instance.latest_version
                instance.update = False
                instance.bind_app(None)
            else:
                not_removed_names.add(p.name)

        if apps_removed:  # updating apps index
            watcher.print(self._i18n['debian.app_index.updating'] + ' ...')
            watcher.change_substatus(self._i18n['debian.app_index.updating'])
            indexed_apps = set(self.app_indexer.read_index())

            if indexed_apps:
                new_index = indexed_apps.difference(apps_removed)
                try:
                    self.app_indexer.update_index(new_index, update_timestamp=False)
                    self._update_apps_index(new_index)
                    self._log.info(f"Debian applications removed from the index: "
                                   f"{', '.join((a.name for a in apps_removed))}")
                except ApplicationIndexError:
                    pass

        watcher.change_substatus('')

        success = True
        if not_removed_names:
            success = pkg.name not in not_removed_names
            not_removed_str = ', '.join((bold(p) for p in sorted(not_removed_names)))
            watcher.show_message(title=self._i18n[f"popup.title.{'warning' if success else 'error'}"],
                                 body=self._i18n['debian.uninstall.failed_to_remove'].format(no=len(not_removed_names),
                                                                                             pkgs=not_removed_str),
                                 type_=MessageType.WARNING if success else MessageType.ERROR)

        return TransactionResult(success=success, installed=None, removed=all_removed)

    def _map_dependents(self, packages_data: Dict[str, Dict[str, object]]) -> Optional[Dict[str, Set[str]]]:
        if packages_data:
            dependents = None
            if packages_data:
                dependents = dict()

                for p, data in packages_data.items():
                    for attr in ('depends', 'predepends'):
                        dep_exps = data.get(attr)

                        if isinstance(dep_exps, tuple):
                            for exp in dep_exps:
                                dep = exp.split(' ')[0].strip()
                                dep_dependents = dependents.get(dep)
                                if dep_dependents is None:
                                    dep_dependents = set()
                                    dependents[dep] = dep_dependents

                                dep_dependents.add(p)

            return dependents

    def get_upgrade_requirements(self, pkgs: List[DebianPackage], root_password: str, watcher: ProcessWatcher) \
            -> UpgradeRequirements:

        transaction = self.aptitude.simulate_upgrade((p.name for p in pkgs))

        if transaction:
            size_to_query = (f'{p.name}={p.latest_version}' for p in (*transaction.to_upgrade,
                                                                      *transaction.to_install,
                                                                      *transaction.to_remove))

            update_extra_attrs = ('compressed size', *(('depends', 'predepends') if transaction.to_install else ()))
            update_data = self.aptitude.show(pkgs=size_to_query, attrs=update_extra_attrs)

            to_install = None
            if transaction.to_install:
                dependents = self._map_dependents(update_data)

                to_install = []
                for p in transaction.to_install:
                    p_data = update_data.get(p.name) if update_data else None
                    req_size = p_data.get('compressed size') if p_data else None

                    reason = None
                    if dependents:
                        p_dependents = dependents.get(p.name)

                        if p_dependents:
                            deps_str = ', '.join((bold(d) for d in p_dependents))
                            reason = self._i18n['debian.transaction.dependency_of'].format(pkgs=deps_str)

                    to_install.append(UpgradeRequirement(pkg=p, reason=reason, required_size=req_size,
                                                         extra_size=p.transaction_size))

                to_install.sort(key=attrgetter('pkg.name'))
                to_install = tuple(to_install)

            to_remove = None
            if transaction.to_remove:
                to_remove = []

                for p in transaction.to_remove:
                    to_remove.append(UpgradeRequirement(pkg=p, required_size=0, extra_size=p.transaction_size))

                to_remove.sort(key=attrgetter('pkg.name'))
                to_remove = tuple(to_remove)

            to_upgrade = None

            if transaction.to_upgrade:
                to_upgrade = []

                for p in transaction.to_upgrade:
                    p_data = update_data.get(p.name) if update_data else None

                    if p_data:
                        req_size = p_data.get('compressed size')
                    else:
                        req_size = None

                    to_upgrade.append(UpgradeRequirement(pkg=p, required_size=req_size, extra_size=p.transaction_size))

                to_upgrade.sort(key=attrgetter('pkg.name'))
                to_upgrade = tuple(to_upgrade)

            return UpgradeRequirements(to_install=to_install, to_upgrade=to_upgrade,
                                       to_remove=to_remove, cannot_upgrade=None)
        else:
            cannot_upgrade = [UpgradeRequirement(pkg=p, reason=self._i18n['error']) for p in pkgs]
            cannot_upgrade.sort(key=attrgetter('pkg.name'))
            return UpgradeRequirements(cannot_upgrade=cannot_upgrade, to_install=None, to_remove=None,
                                       to_upgrade=[])

    def get_managed_types(self) -> Set[Type[SoftwarePackage]]:
        if self._types is None:
            self._types = {DebianPackage}

        return self._types

    def get_info(self, pkg: DebianPackage) -> Optional[dict]:
        info = {'00.name': pkg.name, '01.version': pkg.version,
                "02.description": pkg.description}

        if pkg.installed and pkg.app:
            info['03.exec'] = pkg.app.exe_path

        extra_info = self.aptitude.show((f'{pkg.name}={pkg.version}',), verbose=True)

        if extra_info:
            extra_info = extra_info.get(pkg.name)

            if extra_info:
                ignored_fields = {'package', 'version', 'description'}

                for field, val in extra_info.items():
                    if field not in ignored_fields and not field.startswith('description'):
                        final_val = get_human_size_str(val) if field in self.aptitude.size_attrs else val
                        final_field = f'04.{field}'

                        if final_field not in self._i18n:
                            self._i18n.default[final_field] = field

                        info[final_field] = final_val  # for sorting

        return info

    def get_history(self, pkg: SoftwarePackage) -> PackageHistory:
        pass

    def install(self, pkg: SoftwarePackage, root_password: str, disk_loader: Optional[DiskCacheLoader],
                watcher: ProcessWatcher) -> TransactionResult:

        watcher.change_substatus(self._i18n['debian.simulate_operation'])
        transaction = self.aptitude.simulate_installation((pkg.name, ))

        if transaction is None or not transaction.to_install:
            return TransactionResult.fail()

        if transaction.to_remove or (transaction.to_install and len(transaction.to_install) > 1):
            watcher.change_substatus(self._i18n['debian.transaction.get_data'])

            deps = tuple(p for p in transaction.to_install or () if p.name != pkg.name)
            removal = tuple(p for p in transaction.to_remove or ())
            all_pkgs = [*deps, *removal]

            pkgs_data = self.aptitude.show(pkgs=(f'{d.name}={d.version}' for d in all_pkgs),
                                           attrs=self.install_show_attrs)

            if pkgs_data:
                for p in all_pkgs:
                    fill_show_data(p, pkgs_data.get(p.name))

            if not self.view.confirm_transaction(to_install=deps, removal=removal, watcher=watcher):
                return TransactionResult.fail()

        watcher.change_substatus(self._i18n['debian.installing_pkgs'])
        handler = ProcessHandler(watcher)

        targets = (p.name for p in transaction.to_install)
        with self.output_handler.start(watcher=watcher, targets=targets, action=AptitudeAction.INSTALL) as handle:
            installed, _ = handler.handle_simple(self.aptitude.install(packages=(pkg.name,),
                                                                       root_password=root_password),
                                                 output_handler=handle)

        if installed:
            self._refresh_apps_index(watcher)

            watcher.change_substatus(self._i18n['debian.install.validating'])

            currently_installed = set(self.aptitude.read_installed_names())

            installed_instances = []
            if currently_installed:
                for p in transaction.to_install:
                    instance = p if p != pkg else pkg
                    if instance.name in currently_installed:
                        instance.installed = True
                        instance.bind_app(self.apps_index.get(instance.name))
                        installed_instances.append(instance)

            removed = None

            if transaction.to_remove:
                removed = [p for p in transaction.to_remove if p.name not in currently_installed]

                not_removed = set(transaction.to_remove).difference(removed)

                if not_removed:
                    not_removed_str = ' '.join(p.name for p in not_removed)
                    self._log.warning(f"The following packages were not removed: {not_removed_str}")

            return TransactionResult(installed=installed_instances, removed=removed,
                                     success=bool(installed_instances and pkg in installed_instances))
        else:
            watcher.change_substatus('')
            return TransactionResult.fail()

    def _refresh_apps_index(self, watcher: ProcessWatcher):
        watcher.change_substatus(self._i18n['debian.app_index.checking'])
        self._log.info("Reading the cached Debian applications")
        indexed_apps = self.app_indexer.read_index()
        self._log.info("Mapping the Debian applications")
        current_apps = self.app_mapper.map_executable_applications()

        if current_apps != indexed_apps:
            watcher.print(self._i18n['debian.app_index.updating'] + '...')
            watcher.change_substatus(self._i18n['debian.app_index.updating'])

            try:
                self.app_indexer.update_index(current_apps)
                self._update_apps_index(current_apps)

                if indexed_apps is not None:
                    new_apps = current_apps.difference(indexed_apps)

                    if new_apps:
                        self._log.info(f"Debian applications added to the index: "
                                       f"{','.join((a.name for a in new_apps))}")

            except ApplicationIndexError:
                pass

    def is_enabled(self) -> bool:
        return self._enabled

    def set_enabled(self, enabled: bool):
        self._enabled = enabled

    def can_work(self) -> Tuple[bool, Optional[str]]:
        if not which('aptitude'):
            return False, self._i18n['missing_dep'].format(dep=bold('aptitude'))

        return True, None

    def requires_root(self, action: SoftwareAction, pkg: Optional[SoftwarePackage]) -> bool:
        if action == action.PREPARE:
            deb_config = self.configman.get_config()
            return SynchronizePackages.should_synchronize(deb_config, NullLoggerFactory.logger())

        return action != SoftwareAction.SEARCH

    def prepare(self, task_manager: Optional[TaskManager], root_password: Optional[str],
                internet_available: Optional[bool]):

        deb_config = self.configman.get_config()

        if SynchronizePackages.should_synchronize(deb_config, self._log):
            sync_pkgs = SynchronizePackages(taskman=task_manager, i18n=self._i18n, logger=self._log,
                                            root_password=root_password, aptitude=self.aptitude)
            sync_pkgs.start()

        if DebianSuggestionsDownloader.should_download(deb_config, self._log, only_positive_exp=True):
            self.suggestions_downloader.register_task(task_manager)
            self.suggestions_downloader.start()

        self.index_apps(root_password=root_password, watcher=None, taskman=task_manager,
                        deb_config=deb_config, check_expiration=True)

    def list_updates(self, internet_available: bool) -> List[PackageUpdate]:
        ignored_updates = set()
        fill_ignored_updates = Thread(target=self._fill_ignored_updates, args=(ignored_updates,))
        fill_ignored_updates.start()

        updates = list()
        for name, version in self.aptitude.read_updates():
            if fill_ignored_updates.is_alive():
                fill_ignored_updates.join()

            if name not in ignored_updates:
                updates.append(PackageUpdate(pkg_id=name, name=name, version=version, pkg_type='debian'))

        return updates

    def list_warnings(self, internet_available: bool) -> Optional[List[str]]:
        pass

    def _fill_installed_names(self, output: Set[str]):
        output.update(self.aptitude.read_installed_names())

    def _fill_suggestions(self, output: Dict[str, int]):
        self.suggestions_downloader.register_task(None)
        suggestions = self.suggestions_downloader.read(self.configman.read_config())

        if suggestions:
            output.update(suggestions)

    def list_suggestions(self, limit: int, filter_installed: bool) -> Optional[List[PackageSuggestion]]:
        name_priority = dict()

        fill_suggestions = Thread(target=self._fill_suggestions, args=(name_priority,))
        fill_suggestions.start()

        if filter_installed:
            installed = set()
            fill_installed = Thread(target=self._fill_installed_names, args=(installed, ))
            fill_installed.start()
        else:
            installed, fill_installed = None, None

        fill_suggestions.join()

        if fill_installed:
            fill_installed.join()

        if not name_priority:
            self._log.info("No Debian package suggestions found")
            return []

        self._log.info(f"Found {len(name_priority)} Debian package suggestions")

        to_load = tuple(name_priority.keys()) if not installed else {*name_priority.keys()}.difference(installed)

        if not to_load:
            return []

        suggestions = []

        for pkg in self.aptitude.search_by_name(to_load):
            prio = name_priority.get(pkg.name)

            if prio:
                suggestions.append(PackageSuggestion(package=pkg, priority=prio))

        return suggestions

    def is_default_enabled(self) -> bool:
        return True

    def launch(self, pkg: SoftwarePackage):
        if isinstance(pkg, DebianPackage):
            final_cmd = pkg.app.exe_path.replace('%U', '')
            Popen(final_cmd, shell=True)

    def get_settings(self) -> Optional[Generator[SettingsView, None, None]]:
        config_ = self.configman.get_config()

        purge_opts = [InputOption(label=self._i18n['yes'].capitalize(), value=True),
                      InputOption(label=self._i18n['no'].capitalize(), value=False)]

        purge_current = tuple(o for o in purge_opts if o.value == bool(config_['remove.purge']))[0]
        sel_purge = SingleSelectComponent(id_='remove.purge',
                                          label=self._i18n['debian.config.remove.purge'],
                                          tooltip=self._i18n['debian.config.remove.purge.tip'],
                                          options=purge_opts,
                                          default_option=purge_current,
                                          type_=SelectViewType.RADIO,
                                          max_width=200,
                                          max_per_line=2)

        sources_app = config_.get('pkg_sources.app')

        if isinstance(sources_app, str) and sources_app not in self.known_sources_apps:
            self._log.warning(f"'pkg_sources.app' ({sources_app}) is not supported. A 'None' value will be considered")
            sources_app = None

        lb_source_auto = self._i18n['debian.config.pkg_sources.app.auto']
        source_opts = [InputOption(id_='auto', value=None, label=lb_source_auto)]

        source_opts.extend((InputOption(id_=a, value=a, label=a) for a in self.known_sources_apps if which(a)))

        source_auto_tip = self._i18n['debian.config.pkg_sources.app.tip'].format(auto=f'"{lb_source_auto}"')
        input_sources = SingleSelectComponent(id_='pkg_sources.app',
                                              label=self._i18n['debian.config.pkg_sources.app'],
                                              tooltip=source_auto_tip,
                                              options=source_opts,
                                              default_option=next(o for o in source_opts if o.value == sources_app),
                                              type_=SelectViewType.COMBO,
                                              max_width=200)

        try:
            app_cache_exp = int(config_.get('index_apps.exp', 0))
        except ValueError:
            self._log.error(f"Unexpected value form Debian configuration property 'index_apps.exp': "
                            f"{config_['index_apps.exp']}. Zero (0) will be considered instead.")
            app_cache_exp = 0

        ti_index_apps_exp = TextInputComponent(id_='index_apps.exp',
                                               label=self._i18n['debian.config.index_apps.exp'],
                                               tooltip=self._i18n['debian.config.index_apps.exp.tip'],
                                               value=str(app_cache_exp), only_int=True,
                                               max_width=60)

        try:
            sync_pkgs_time = int(config_.get('sync_pkgs.time', 0))
        except ValueError:
            self._log.error(f"Unexpected value form Debian configuration property 'sync_pkgs.time': {config_['sync_pkgs.time']}. "
                            f"Zero (0) will be considered instead.")
            sync_pkgs_time = 0

        ti_sync_pkgs = TextInputComponent(id_='sync_pkgs.time',
                                          label=self._i18n['debian.config.sync_pkgs.time'],
                                          tooltip=self._i18n['debian.config.sync_pkgs.time.tip'],
                                          value=str(sync_pkgs_time), only_int=True,
                                          max_width=60)

        try:
            suggestions_exp = int(config_.get('suggestions.exp', 0))
        except ValueError:
            self._log.error(f"Unexpected value form Debian configuration property 'suggestions.exp': {config_['suggestions.exp']}. "
                            f"Zero (0) will be considered instead.")
            suggestions_exp = 0

        ti_suggestions_exp = TextInputComponent(id_='suggestions.exp',
                                                label=self._i18n['debian.config.suggestions.exp'],
                                                tooltip=self._i18n['debian.config.suggestions.exp.tip'],
                                                value=str(suggestions_exp), only_int=True,
                                                max_width=60)

        panel = PanelComponent([FormComponent([input_sources, sel_purge, ti_sync_pkgs, ti_index_apps_exp,
                                               ti_suggestions_exp])])
        yield SettingsView(self, panel)

    def save_settings(self, component: PanelComponent) -> Tuple[bool, Optional[List[str]]]:
        config_ = self.configman.get_config()

        container = component.get_component_by_idx(0, FormComponent)

        for prop, type_ in {'remove.purge': SingleSelectComponent,
                            'pkg_sources.app': SingleSelectComponent,
                            'index_apps.exp': TextInputComponent,
                            'sync_pkgs.time': TextInputComponent,
                            'suggestions.exp': TextInputComponent}.items():
            comp = container.get_component(prop, type_)

            val = None
            if isinstance(comp, SingleSelectComponent):
                val = comp.get_selected()
            elif isinstance(comp, TextInputComponent):
                val = comp.get_int_value()

            config_[prop] = val

        try:
            self.configman.save_config(config_)
            return True, None
        except:
            return False, [traceback.format_exc()]

    def gen_custom_actions(self) -> Generator[CustomSoftwareAction, None, None]:
        if self._default_actions is None:
            self._default_actions = (CustomSoftwareAction(i18n_label_key='debian.action.sync_pkgs',
                                                          i18n_status_key='debian.task.sync_pkgs.status',
                                                          i18n_description_key='debian.action.sync_pkgs.desc',
                                                          icon_path=DEBIAN_ICON_PATH,
                                                          manager_method='synchronize_packages',
                                                          requires_root=True),
                                     CustomSoftwareAction(i18n_label_key='debian.action.index_apps',
                                                          i18n_status_key='debian.task.app_index.status',
                                                          i18n_description_key='debian.action.index_apps.desc',
                                                          icon_path=DEBIAN_ICON_PATH,
                                                          manager_method='index_apps',
                                                          requires_root=False)
                                     )

        yield from self._default_actions

        for _ in self.get_installed_source_apps():
            yield self.action_launch_sources
            break

    def _write_ignored_updates(self, packages: Iterable[str]):
        Path(os.path.dirname(self.file_ignored_updates)).mkdir(parents=True, exist_ok=True)

        with open(self.file_ignored_updates, 'w+') as f:
            f.write('\n'.join(n for n in sorted(packages)))

    def ignore_update(self, pkg: DebianPackage):
        ignored_packages = set()
        self._fill_ignored_updates(ignored_packages)

        if pkg.name not in ignored_packages:
            pkg.updates_ignored = True
            ignored_packages.add(pkg.name)
            self._write_ignored_updates(ignored_packages)

    def revert_ignored_update(self, pkg: DebianPackage):
        ignored_packages = set()
        self._fill_ignored_updates(ignored_packages)

        if pkg.name in ignored_packages:
            pkg.updates_ignored = False
            ignored_packages.remove(pkg.name)
            self._write_ignored_updates(ignored_packages)

    def synchronize_packages(self, root_password: Optional[str], watcher: ProcessWatcher) -> bool:
        return SynchronizePackages(i18n=self._i18n, logger=self._log, root_password=root_password, watcher=watcher,
                                   aptitude=self.aptitude, taskman=TaskManager()).run()

    def index_apps(self, root_password: Optional[str], watcher: Optional[ProcessWatcher],
                   taskman: Optional[TaskManager] = None, deb_config: Optional[dict] = None,
                   check_expiration: bool = False) -> bool:
        _config = deb_config if deb_config else self.configman.get_config()
        _taskman = taskman if taskman else TaskManager()

        map_apps = MapApplications(taskman=_taskman, app_indexer=self.app_indexer,
                                   i18n=self._i18n, logger=self._log, deb_config=deb_config,
                                   app_mapper=self.app_mapper, check_expiration=check_expiration,
                                   watcher=watcher)
        map_apps.start()

        gen_app_index = UpdateApplicationIndex(taskman=_taskman, app_indexer=self.app_indexer,
                                               i18n=self._i18n, logger=self._log,
                                               mapping_apps=map_apps)
        gen_app_index.start()

        map_apps.join()
        self._update_apps_index(map_apps.apps)
        gen_app_index.join()
        return True

    def purge(self, pkg: DebianPackage, root_password: str, watcher: ProcessWatcher) -> bool:
        if not self.view.confirm_purge(pkg.name, watcher):
            return False

        res = self.uninstall(pkg=pkg, root_password=root_password, watcher=watcher, disk_loader=None, purge=True)
        return res.success

    def launch_sources_app(self, root_password: str, watcher: ProcessWatcher) -> bool:
        deb_config = self.configman.get_config()

        sources_app = deb_config.get('pkg_sources.app')

        if isinstance(sources_app, str) and sources_app not in self.known_sources_apps:
            watcher.show_message(title=self._i18n['popup.title.error'],
                                 body=self._i18n['debian.action.sources.unsupported'].format(app=bold(sources_app)))
            return False

        if sources_app:
            if not which(sources_app):
                watcher.show_message(title=self._i18n['popup.title.error'],
                                     body=self._i18n['debian.action.sources.not_installed'],
                                     type_=MessageType.ERROR)
                return False

            Popen(sources_app, shell=True)
            return True

        for app in self.get_installed_source_apps():
            Popen(app, shell=True)
            return True

        watcher.show_message(title=self._i18n['popup.title.error'],
                             body=self._i18n['debian.action.sources.not_installed'],
                             type_=MessageType.ERROR)
        return False

    @property
    def app_indexer(self) -> ApplicationIndexer:
        if self._app_indexer is None:
            self._app_indexer = ApplicationIndexer(self._log)

        return self._app_indexer

    @property
    def apps_index(self) -> Dict[str, DebianApplication]:
        if self._apps_index is None:
            self._update_apps_index(self.app_indexer.read_index())

        return self._apps_index

    @property
    def install_show_attrs(self) -> Set[str]:
        if self._install_show_attrs is None:
            self._install_show_attrs = {'description', 'maintainer', 'section', 'compressed size'}

        return self._install_show_attrs

    @property
    def configman(self) -> DebianConfigManager:
        if self._configman is None:
            self._configman = DebianConfigManager()

        return self._configman

    @property
    def view(self) -> DebianViewBridge:
        if self._view is None:
            self._view = DebianViewBridge(screen_width=self.context.screen_width,
                                          screen_heigth=self.context.screen_height,
                                          i18n=self._i18n)

        return self._view

    @property
    def app_mapper(self) -> ApplicationsMapper:
        if self._app_mapper is None:
            self._app_mapper = ApplicationsMapper(logger=self._log)

        return self._app_mapper

    @property
    def aptitude(self) -> Aptitude:
        if self._aptitude is None:
            self._aptitude = Aptitude(self._log)

        return self._aptitude

    @property
    def output_handler(self) -> AptitudeOutputHandlerFactory:
        if self._output_handler is None:
            self._output_handler = AptitudeOutputHandlerFactory(i18n=self._i18n)

        return self._output_handler

    @property
    def action_launch_sources(self) -> CustomSoftwareAction:
        if self._action_launch_sources is None:
            self._action_launch_sources = CustomSoftwareAction(i18n_label_key='debian.action.sources',
                                                               i18n_status_key='debian.task.sources.status',
                                                               i18n_description_key='debian.action.sources.desc',
                                                               icon_path=DEBIAN_ICON_PATH,
                                                               manager_method='launch_sources_app',
                                                               requires_confirmation=False,
                                                               requires_root=False)

        return self._action_launch_sources

    @property
    def known_sources_apps(self) -> Tuple[str, ...]:
        if self._known_sources_apps is None:
            self._known_sources_apps = ('software-properties-gtk',)

        return self._known_sources_apps

    @property
    def file_ignored_updates(self) -> str:
        if self._file_ignored_updates is None:
            self._file_ignored_updates = f'{CONFIG_DIR}/debian/updates_ignored.txt'

        return self._file_ignored_updates

    @property
    def suggestions_downloader(self) -> DebianSuggestionsDownloader:
        if not self._suggestions_downloader:
            self._suggestions_downloader = DebianSuggestionsDownloader(i18n=self._i18n, logger=self._log,
                                                                       http_client=self.context.http_client)

        return self._suggestions_downloader

    def get_installed_source_apps(self) -> Generator[str, None, None]:
        for app in self.known_sources_apps:
            if shutil.which(app):
                yield app



import os
import traceback
from datetime import datetime, timedelta
from logging import Logger
from pathlib import Path
from threading import Thread
from typing import Optional, Dict

from bauh.api.abstract.handler import TaskManager
from bauh.api.abstract.model import SuggestionPriority
from bauh.api.http import HttpClient
from bauh.gems.debian import DEBIAN_ICON_PATH, DEBIAN_CACHE_DIR
from bauh.view.util.translation import I18n


class DebianSuggestionsDownloader(Thread):

    _file_suggestions: Optional[str] = None
    _file_suggestions_ts: Optional[str] = None
    _url_suggestions: Optional[str] = None

    @classmethod
    def file_suggestions(cls) -> str:
        if cls._file_suggestions is None:
            cls._file_suggestions = f'{DEBIAN_CACHE_DIR}/suggestions.txt'

        return cls._file_suggestions

    @classmethod
    def file_suggestions_timestamp(cls) -> str:
        if cls._file_suggestions_ts is None:
            cls._file_suggestions_ts = f'{cls.file_suggestions()}.ts'

        return cls._file_suggestions_ts

    @classmethod
    def url_suggestions(cls) -> str:
        if cls._url_suggestions is None:
            cls._url_suggestions = 'https://raw.githubusercontent.com/vinifmor/bauh-files' \
                                   '/master/debian/suggestions.txt'

        return cls._url_suggestions

    def __init__(self, logger: Logger, http_client: HttpClient, i18n: I18n):
        super(DebianSuggestionsDownloader, self).__init__()
        self._log = logger
        self.i18n = i18n
        self.http_client = http_client
        self._taskman: Optional[TaskManager] = None
        self.task_id = 'debian.suggs'

    def register_task(self, taskman: Optional[TaskManager]):
        self._taskman = taskman
        if taskman:
            self._taskman.register_task(id_=self.task_id, label=self.i18n['task.download_suggestions'],
                                        icon_path=DEBIAN_ICON_PATH)

    @property
    def taskman(self) -> TaskManager:
        if self._taskman is None:
            self._taskman = TaskManager()

        return self._taskman

    @classmethod
    def should_download(cls, debian_config: dict, logger: Logger, only_positive_exp: bool = False) -> bool:
        try:
            exp_hours = int(debian_config['suggestions.exp'])
        except ValueError:
            logger.error(f"The Debian configuration property 'suggestions.expiration' has a non int value set: "
                         f"{debian_config['suggestions']['expiration']}")
            return not only_positive_exp

        if exp_hours <= 0:
            logger.info("Suggestions cache is disabled")
            return not only_positive_exp

        if not os.path.exists(cls.file_suggestions()):
            logger.info(f"'{cls.file_suggestions()}' not found. It must be downloaded")
            return True

        if not os.path.exists(cls.file_suggestions()):
            logger.info(f"'{cls.file_suggestions()}' not found. The suggestions file must be downloaded.")
            return True

        with open(cls.file_suggestions_timestamp()) as f:
            timestamp_str = f.read()

        try:
            suggestions_timestamp = datetime.fromtimestamp(float(timestamp_str))
        except:
            logger.error(f'Could not parse the Debian cached suggestions timestamp: {timestamp_str}')
            traceback.print_exc()
            return True

        update = suggestions_timestamp + timedelta(hours=exp_hours) <= datetime.utcnow()
        return update

    def _save(self, text: str, timestamp: float):
        self._log.info(f"Caching suggestions to '{self.file_suggestions()}'")

        cache_dir = os.path.dirname(self.file_suggestions())

        try:
            Path(cache_dir).mkdir(parents=True, exist_ok=True)
            cache_dir_ok = True
        except OSError:
            self._log.error(f"Could not create cache directory '{cache_dir}'")
            traceback.print_exc()
            cache_dir_ok = False

        if cache_dir_ok:
            try:
                with open(self.file_suggestions(), 'w+') as f:
                    f.write(text)
            except:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions()}'")
                traceback.print_exc()

            try:
                with open(self.file_suggestions_timestamp(), 'w+') as f:
                    f.write(str(timestamp))
            except:
                self._log.error(f"An exception happened while writing the file '{self.file_suggestions_timestamp()}'")
                traceback.print_exc()

    def parse_suggestions(self, suggestions_str: str) -> Dict[str, SuggestionPriority]:
        output = dict()
        for line in suggestions_str.split('\n'):
            clean_line = line.strip()

            if clean_line:
                line_split = clean_line.split(':', 1)

                if len(line_split) == 2:
                    try:
                        prio = int(line_split[0])
                    except ValueError:
                        self._log.warning(f"Could not parse Debian package suggestion: {line}")
                        continue

                    output[line_split[1]] = SuggestionPriority(prio)

        return output

    def read_cached(self) -> Optional[Dict[str, SuggestionPriority]]:
        self._log.info(f"Reading cached suggestions file '{self.file_suggestions()}'")

        try:
            with open(self.file_suggestions()) as f:
                sugs_str = f.read()
        except FileNotFoundError:
            self._log.warning(f"Cached suggestions file does not exist ({self.file_suggestions()})")
            return

        if not sugs_str:
            self._log.warning(f"Cached suggestions file '{self.file_suggestions()}' is empty")
            return

        return self.parse_suggestions(sugs_str)

    def download(self) -> Optional[Dict[str, SuggestionPriority]]:
        self.taskman.update_progress(self.task_id, progress=1, substatus=None)

        self._log.info(f"Downloading suggestions from {self.url_suggestions()}")
        res = self.http_client.get(self.url_suggestions())

        suggestions = None
        if res.status_code == 200 and res.text:
            self.taskman.update_progress(self.task_id, progress=50, substatus=None)
            suggestions = self.parse_suggestions(res.text)

            if suggestions:
                self._save(text=res.text, timestamp=datetime.utcnow().timestamp())
            else:
                self._log.warning("No Debian suggestions to cache")
        else:
            self._log.warning(f"Could not retrieve Debian suggestions. "
                              f"Response (status={res.status_code}, text={res.text})")

        self.taskman.update_progress(self.task_id, progress=100, substatus=None)
        self.taskman.finish_task(self.task_id)
        return suggestions

    def read(self, debian_config: dict) -> Optional[Dict[str, int]]:
        if self.should_download(debian_config=debian_config, logger=self._log):
            return self.download()

        return self.read_cached()

    def run(self):
        self.download()
